# -*- coding: utf-8 -*-

"""
Utility functions
"""

import re
import os, sys
import logging
import nltk
import nltkor
import _pickle
import chardet
#from nltkor import Kor_char
from nltkor.tokenize import Ko_tokenize
from nltkor.tag.libs import config
import numpy as np

#from nltk.tokenize.regexp import RegexpTokenizer
#from nltk.tokenize import TreebankWordTokenizer
from . import attributes


def get_word_from_morph_lexicon(root, word, tags, space_flag):
	'''
	space_flag: if True then including space, otherwise do not including space
	'''

	values = list()
	value_data = list()
	if not word: return root.keys()

	current_dict = root
	_end = '$$'
	s = 0
	for i, letter in enumerate(word):
		#print(i, '>', letter, current_dict)
		if letter in current_dict:
			#print(letter, current_dict[letter])
			current_dict = current_dict[letter]
			if _end in current_dict :
				for idx in range(i-s): values.pop()
				values.append(word[s:i+1])
				for idx in range(i-s): value_data.pop()
				value_data.append(current_dict[_end])
			else: values.append(letter); value_data.append(tags[i])
		else:
			#print('==', letter, values)
			if space_flag or letter != ' ':
				values.append(letter)				# 최장일치 : -1
				value_data.append(tags[i])
			s = i+1
			current_dict = root
	else:
		if values: return values, value_data
		else:	return list(word), tags

def intersperse(lst, item):
	result = [item] * (len(lst) * 2 - 1)
	result[0::2] = lst
	return result

def get_word(root, word, tags, space_flag=False) :
	'''
	space_flag : True : 공백이 있어도 매칭됨
							 False: 공백이 있으면 매칭 안됨
	'''
	word_list = get_word_from_morph_lexicon(root, word, tags, space_flag)
	return word_list



def tokenize(text, use_sent_tokenizer=True):
		"""
		Call the tokenizer function for the given language.
		The returned tokens are in a list of lists, one for each sentence.

		:param use_sent_tokenizer: True : use sentence tokenizer
									False : sentence per line
		"""
		return tokenize_ko(text, use_sent_tokenizer)

def tokenize_ko(text, use_sent_tokenizer=True, clean=True):
		"""
		text: string
		Return a list of lists of the tokens in text, separated by sentences.
		"""
		if clean:
			text = clean_kotext(text)

		if use_sent_tokenizer:
			## False: 띄어쓰기 무시, True: 띄어쓰기 고려
			sentences = [Ko_tokenize.syllable(sentence, True) for sentence in Ko_tokenize.sentence(text)]
		else:
			sentences = [Ko_tokenize.syllable(text, True)]

		return sentences

def clean_kotext(text, correct=False):
		"""
		1. 특수 공백문자를 공백으로 처리
		Apply some transformations to the text, such as
		replacing digits for 9 and simplifying quotation marks.

		:param correct: If True, tries to correct punctuation misspellings.
		"""
		# replaces different kinds of quotation marks with "
		# take care not to remove apostrophes
		'''
		text = re.sub(r"(?u)(^|\W)[‘’′`']", r'\1"', text)
		text = re.sub(r"(?u)[‘’`′'](\W|$)", r'"\1', text)
		text = re.sub(r'(?u)[«»“”]', '"', text)

		if correct:
				# tries to fix mistyped tokens (common in Wikipedia-pt) as ,, '' ..
				text = re.sub(r'(?<!\.)\.\.(?!\.)', '.', text) # take care with ellipses
				text = re.sub(r'([,";:])\1,', r'\1', text)

				# inserts space after leading hyphen. It happens sometimes in cases like:
				# blablabla -that is, bloblobloblo
				text = re.sub(' -(?=[^\W\d_])', ' - ', text)
		'''

		# replaces numbers with the 9's
		text = re.sub(r'\xa0', ' ', text)
		text = re.sub(u'　', ' ', text)
		text = re.sub(u' ', ' ', text)
		text = re.sub(u' +', ' ', text) # 연속된 공백 처리
		# replaces numbers with the 9's, 사전쪽과 같이 판단할 것
		#text = re.sub(r'\d', '9', text)
		# replaces english character with the a's
		#text = re.sub(r'[a-zA-Z]', 'a', text)
		# replaces chinses character with the 家's
		#for x in re.findall(r'[\u4e00-\u9fff]', text):
		#	text = re.sub(x, '家', text)

		# replaces special ellipsis character
		#text = text.replace(u'…', '...')

		return text


def generate_feature_vectors(num_vectors, num_features, min_value=-0.1, max_value=0.1):
		"""
		Generates vectors of real numbers, to be used as word features.
		Vectors are initialized randomly. Returns a 2-dim numpy array.
		"""
		logger = logging.getLogger("Logger")
		#table = (max_value * 2) * np.random.random((num_vectors, num_vectors, num_features, num_features)) + min_value
		table = (max_value * 2) * np.random.random((num_vectors, num_features)) + min_value
		logger.debug("Generated %d feature vectors with %d features each." % (num_vectors, num_features))
		print("Generated %d feature vectors with %d features each." % (num_vectors, num_features))

		return table


def count_lines(filename):
		"""Counts and returns how many non empty lines in a file there are."""
		with open(filename, 'r') as f:
				lines = [x for x in list(f) if x.strip()]
		return len(lines)

def _create_affix_tables(affix, table_list, num_features):
		"""
		Internal helper function for loading suffix or prefix feature tables
		into the given list.
		affix should be either 'suffix' or 'prefix'.
		"""
		logger = logging.getLogger('Logger')
		logger.info('Generating %s features...' % affix)
		tensor = []
		codes = getattr(attributes.Affix, '%s_codes' % affix)
		num_affixes_per_size = getattr(attributes.Affix, 'num_%ses_per_size' % affix)
		for size in codes:

				# use num_*_per_size because it accounts for special suffix codes
				num_affixes = num_affixes_per_size[size]
				table = generate_feature_vectors(num_affixes, num_features)
				tensor.append(table)

		# affix attribute actually has a 3-dim tensor
		# (concatenation of 2d tables, one for each suffix size)
		for table in tensor:
				table_list.append(table)

def create_feature_tables(args, md, text_reader):
		"""
		Create the feature tables to be used by the network. If the args object
		contains the load_features option as true, the feature table for word types
		is loaded instead of being created. The actual number of
		feature tables will depend on the argument options.

		:param arguments: Parameters supplied to the program
		:param md: metadata about the network
		:param text_reader: The TextReader being used.
		:returns: all the feature tables to be used
		"""

		logger = logging.getLogger("Logger")
		feature_tables = []

		if not args.load_types:
				logger.info("Generating word type features...")
				table_size = len(text_reader.word_dict)
				types_table = generate_feature_vectors(table_size, args.num_features)
		else:
				logger.info("Loading word type features...")
				# check if there is a word feature file specific for the task
				# if not, load a generic one
				filename = md.paths[md.type_features]
				if os.path.exists(filename):
						types_table = load_features_from_file(filename)
				else:
						filename = md.paths['type_features']
						types_table = load_features_from_file(filename)

				if len(types_table) < len(text_reader.word_dict):
						# the type dictionary provided has more types than
						# the number of feature vectors. So, let's generate
						# feature vectors for the new types by replicating the vector
						# associated with the RARE word
						diff = len(text_reader.word_dict) - len(types_table)
						logger.warning("Number of types in feature table and dictionary differ.")
						logger.warning("Generating features for %d new types." % diff)
						num_features = len(types_table[0])
						new_vecs =	generate_feature_vectors(diff, num_features)
						types_table = np.append(types_table, new_vecs, axis=0)

				elif len(types_table) < len(text_reader.word_dict):
						logger.warning("Number of features provided is greater than the number of tokens\
						in the dictionary. The extra features will be ignored.")

		feature_tables.append(types_table)	# head
		#print(md.task)
		#if md.task in ['labeled_dependency', 'unlabeled_dependency']:
		#	feature_tables.append(types_table)	# tail

		# Capitalization
		if md.use_caps:
				logger.info("Generating capitalization features...")
				caps_table = generate_feature_vectors(attributes.Caps.num_values, args.caps)
				feature_tables.append(caps_table)

		# Prefixes
		if md.use_prefix:
				_create_affix_tables('prefix', feature_tables, args.prefix)

		# Suffixes
		if md.use_suffix:
				_create_affix_tables('suffix', feature_tables, args.suffix)

		# POS tags
		if md.use_pos:
				logger.info("Generating POS features...")
				num_pos_tags = text_reader.get_num_pos_tags()
				pos_table = generate_feature_vectors(num_pos_tags, args.pos)
				#feature_tables.append(pos_table) # head    # 여기와 *_reader의 converter와 일치해야 한다.
				feature_tables.append(pos_table) # tail

		# chunk tags
		if md.use_chunk:
				logger.info("Generating chunk features...")
				num_chunk_tags = count_lines(md.paths['chunk_tags'])
				chunk_table = generate_feature_vectors(num_chunk_tags, args.chunk)
				feature_tables.append(chunk_table)

		#print(len(feature_tables))
		return feature_tables



def set_distance_features(max_dist=None,
													num_target_features=None, num_pred_features=None):
		"""
		Returns the distance feature tables to be used by a convolutional network.
		One table is for relative distance to the target predicate, the other
		to the predicate.

		:param max_dist: maximum distance to be used in new vectors.
		"""
		logger = logging.getLogger("Logger")

		# max_dist before/after, 0 distance, and distances above the max
		max_dist = 2 * (max_dist + 1) + 1
		logger.info("Generating target word distance features...")
		target_dist = generate_feature_vectors(max_dist, num_target_features)
		logger.info("Generating predicate distance features...")
		pred_dist = generate_feature_vectors(max_dist, num_pred_features)

		return [target_dist, pred_dist]


def set_logger(level):
		"""Sets the logger to be used throughout the system."""
		log_format = '%(message)s'
		logging.basicConfig(format=log_format)
		logger = logging.getLogger("Logger")
		logger.setLevel(level)

def load_features_from_file(features_file):
		"""Reads a file with features written as binary data."""
		return np.load(features_file)

def save_features_to_file(table, features_file):
		"""Saves a feature table to a given file, writing binary data."""
		np.save(features_file, table)

def convert_iobes_to_bracket(tag):
		"""
		Convert tags from the IOBES scheme to the CoNLL bracketing.

		Example:
		B-A0 -> (A0*
		I-A0 -> *
		E-A0 -> *)
		S-A1 -> (A1*)
		O		-> *
		"""
		if tag.startswith('I') or tag.startswith('O'):
				return '*'
		if tag.startswith('B'):
				return '(%s*' % tag[2:]
		if tag.startswith('E'):
				return '*)'
		if tag.startswith('S'):
				return '(%s*)' % tag[2:]
		else:
				raise ValueError("Unknown tag: %s" % tag)

def boundaries_to_arg_limits(boundaries):
		"""
		Converts a sequence of IOBES tags delimiting arguments to an array
		of argument boundaries, used by the network.
		"""
		limits = []
		start = None

		for i, tag in enumerate(boundaries):
				if tag == 'S':
						limits.append([i, i])
				elif tag == 'B':
						start = i
				elif tag == 'E':
						limits.append([start, i])

		return np.array(limits, np.int)

class PickleConverter:
	"""txt 확장자 사전을 pickle 확장자로 변환"""
	def __init__(self, morph_pickle=None, co_pickle=None, prob_pickle=None):
		self.morph_pickle = morph_pickle or config.FILES.get("pos_morph_lexicon")

	def _convert_morph_lexicon(self, root, word, data):
		'''
		root = dict()
		...
		_convert_morph_dict(root, u_key, u_data)
		'''
		current_dict = root
		_end = '$$'
		for letter in word:
			current_dict = current_dict.setdefault(letter, {})
		current_dict = current_dict.setdefault(_end, data)
		return root	

	def convert_morph_lexicon(self, filename=None):
		filename = filename if filename else self.morph_pickle
		txt_filename = filename.replace('.pickle', '.txt')
		morph_dict = {} 

		with open(txt_filename, 'rb') as f:
			raw_data = f.read(1024)
			detected = chardet.detect(raw_data).get('encoding', 'utf-8')
		with open(txt_filename, 'rt', encoding=detected) as f:
			for line in f:
				if ';;' in line[:2]: continue
				try:
					k, v = line.strip().split('\t')
				except:
					print('morph lexicon error : ', line)
				self._convert_morph_lexicon(morph_dict, k, v)

		with open(filename, 'wb') as f:
			_pickle.dump(morph_dict, f, 2)



# -*- coding: utf-8 -*-
# Natural Language Toolkit: Taggers
#
# Copyright (C) 2001-2020 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
#         Steven Bird <stevenbird1@gmail.com> (minor additions)
# URL: <http://nltk.org/>
# For license information, see LICENSE.TXT
"""
NLTK Taggers

This package contains classes and interfaces for part-of-speech
tagging, or simply "tagging".

A "tag" is a case-sensitive string that specifies some property of a token,
such as its part of speech.  Tagged tokens are encoded as tuples
``(tag, token)``.  For example, the following tagged token combines
the word ``'fly'`` with a noun part of speech tag (``'NN'``):

    >>> tagged_tok = ('fly', 'NN')

An off-the-shelf tagger is available for English. It uses the Penn Treebank tagset:

    >>> from nltk import pos_tag, word_tokenize
    >>> pos_tag(word_tokenize("John's big idea isn't all that bad."))
    [('John', 'NNP'), ("'s", 'POS'), ('big', 'JJ'), ('idea', 'NN'), ('is', 'VBZ'),
    ("n't", 'RB'), ('all', 'PDT'), ('that', 'DT'), ('bad', 'JJ'), ('.', '.')]

A Russian tagger is also available if you specify lang="rus". It uses
the Russian National Corpus tagset:

    >>> pos_tag(word_tokenize("Илья оторопел и дважды перечитал бумажку."), lang='rus')    # doctest: +SKIP
    [('Илья', 'S'), ('оторопел', 'V'), ('и', 'CONJ'), ('дважды', 'ADV'), ('перечитал', 'V'),
    ('бумажку', 'S'), ('.', 'NONLEX')]

This package defines several taggers, which take a list of tokens,
assign a tag to each one, and return the resulting list of tagged tokens.
Most of the taggers are built automatically based on a training corpus.
For example, the unigram tagger tags each word *w* by checking what
the most frequent tag for *w* was in a training corpus:

    >>> from nltk.corpus import brown
    >>> from nltk.tag import UnigramTagger
    >>> tagger = UnigramTagger(brown.tagged_sents(categories='news')[:500])
    >>> sent = ['Mitchell', 'decried', 'the', 'high', 'rate', 'of', 'unemployment']
    >>> for word, tag in tagger.tag(sent):
    ...     print(word, '->', tag)
    Mitchell -> NP
    decried -> None
    the -> AT
    high -> JJ
    rate -> NN
    of -> IN
    unemployment -> None

Note that words that the tagger has not seen during training receive a tag
of ``None``.

We evaluate a tagger on data that was not seen during training:

    >>> tagger.evaluate(brown.tagged_sents(categories='news')[500:600])
    0.73...

For more information, please consult chapter 5 of the NLTK Book.
"""


from nltkor.tag.espresso_tag import EspressoTagger
#import nltkor.tag
from nltkor.tag.libs import taggers
from .libs import PickleConverter


from xml.etree.ElementTree import parse
import os, re
from operator import eq
import time
import nltkor
from nltkor.sejong.sejong_download import SejongDir


common_path=os.path.dirname(nltkor.sejong.__file__)
class Entry():

    def __init__(self, name, en, pos):
        self.name = name
        self.entry = en
        self.pos = pos
        self.SejongDir = SejongDir()
        

    def __repr__(self):
        return "%s('%s')" % (type(self).__name__, self.name)



    # sense객체 리턴
    def senses(self):
        list = []

        allsense = self.entry.findall("sense")
        for se in allsense:
            try:
                ss = str(self.name + "." + se.attrib['n'])
            except KeyError:
                ss = str(self.name)
            temp = Sense(ss, se, self.pos)
            list.append(temp)

        return list

    # 숙어
    def idm(self):
        list = []
        try:
            id = self.entry.find("idm_grp")
            idm = id.findall("idm")
        except AttributeError:
            return list

        for tmp in idm:
            if tmp.text is None:
                return list

            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
            list.append(tmp.text)

        return list

    # 복합어
    def comp(self):

        list = []
        try:
            mor = self.entry.find("morph_grp")
            comp = mor.findall("comp")
        except AttributeError:
            return list

        for tmp in comp:
            if tmp.text is None:
                return list

            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
            list.append(tmp.text)

        return list

    # 파생어
    def der(self):
        list = []
        try:
            mor = self.entry.find("morph_grp")
            comp = mor.findall("der")
        except AttributeError:
            return list
        for tmp in comp:
            if tmp.text is None:
                return list

            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
            list.append(tmp.text)

        return list


class Sense():

    def __init__(self, name, se, pos):
        self.name = name
        self.sense = se
        self.pos = pos

    def __repr__(self):
        return "%s('%s')" % (type(self).__name__, self.name)

    # 공통 태그
    def common_lr(self, sense):
        sem = sense.find("sem_grp")
        lr = sem.find("lr")
        return lr

        # sem
    #sem
    def sem(self):
        list = []
        sem = self.sense.find("sem_grp")
        synn = sem.find("sem_class")
        try:
            synn = synn.text
        except AttributeError:
            return list

        list.append(synn)
        return list

        # if None in list:
        #     list = []
        #     return list
        # else:
        #     return list
    # 동의어
    def syn(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            synn = lr.findall("syn")
        except AttributeError:
            return list

        for tmp in synn:
            list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 반의어
    def ant(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            ant = lr.findall("ant")
        except AttributeError:
            return list

        for tmp in ant:
            list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 동위어
    def coord(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            coo = lr.findall("coord")
        except AttributeError:
            return list

        for tmp in coo:
            list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 부분어
    def mero(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            me = lr.findall("mero")
        except AttributeError:
            return list

        for tmp in me:
            list.append(tmp.text)

        '''if not list:
            return("@@@@@",list)
        '''
        if None in list:
            list = []
            return list
        else:
            return list
    # 상위어
    def hyper(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            hy = lr.findall("hyper")
        except AttributeError:
            return list

        for tmp in hy:
            list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 하위어
    def hypo(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            hy = lr.findall("hypo")
        except AttributeError:
            return list

        for tmp in hy:
            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
                list.append(tmp.text)
            else:
                list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 전체어
    def holo(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            ho = lr.findall("holo")
        except AttributeError:
            return list

        for tmp in ho:
            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
                list.append(tmp.text)
            else:
                list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 관련어
    def rel(self):
        list = []
        lr = self.common_lr(self.sense)
        try:
            rel = lr.findall("rel")
        except AttributeError:
            return list

        for tmp in rel:
            list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 예시
    def example(self):
        list = []

        if self.pos != 'nng_s':
            return list

        else:
            sem = self.sense.find("sem_grp")
            eg = sem.findall("eg")
            for tmp in eg:
                if '~' in tmp.text:
                    name = self.name.split('.')
                    tmp.text = tmp.text.replace('~', name[0])
                    list.append(tmp.text)
                else:
                    list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 영어
    def trans(self):
        list = []
        sem = self.sense.find("sem_grp")
        trs = sem.findall("trans")
        for tmp in trs:
            list.append(tmp.text)

        if None in list:
            list = []
            return list
        else:
            return list
    # 형용사 결합
    def comb_aj(self):
        list = []

        try:
            syn = self.sense.find("syn_grp")
            aj = syn.findall("comb_aj")
        except AttributeError:
            return list

        for tmp in aj:
            if tmp.text is None:
                return list

            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
            list.append(tmp.text)

        return list
    # 명사 결합
    def comb_n(self):
        list = []
        try:
            syn = self.sense.find("syn_grp")
            n = syn.findall("comb_n")
        except AttributeError:
            return list
        for tmp in n:
            if tmp.text is None:
                return list

            if '~' in tmp.text:
                name = self.name.split('.')
                tmp.text = tmp.text.replace('~', name[0])
            list.append(tmp.text)

        return list
    # 동사 결합
    def comb_v(self):
        list = []
        try:
            syn = self.sense.find("syn_grp")
            v = syn.findall("comb_v")
        except AttributeError:
            return list

        for tmp in v:
            v = tmp.find("form").text
            if v is None:
                return list

            if '~' in v:
                name = self.name.split('.')
                v = v.replace('~', name[0])
            list.append(v)
        return list

    # frame
    def sel_rst(self):

        final = {}
        list = []

        if self.pos == 'nng_s':
            return list

        frame_grps = self.sense.findall("frame_grp")

        for grp in frame_grps:  # 각각의 frame_grp type
            sub_list = []
            for subsense in grp.findall('subsense'):  # n개의 subsense
                str = ""
                eg_list = []
                check = 0
                for sel_rst in subsense.findall('sel_rst'):  # m개의 sel_rst
                    check += 1
                    for tmp in sel_rst.attrib.items():

                        if (tmp[0] == 'arg'):
                            str += ("<" + tmp[0] + "=" + tmp[1] + " ")

                        if (tmp[0] == 'tht'):
                            str += (tmp[0] + "=" + tmp[1] + ">")
                    try:
                        str += (sel_rst.text)
                    except TypeError:
                        str += ' '

                    if (check != len(subsense.findall('sel_rst'))):
                        str += ', '

                for eg in subsense.findall('eg'):
                    eg_list.append(eg.text)

                sub_list.append(str)
                sub_list.append(eg_list)

            final[grp.find('frame').text] = sub_list

        return final

    # 최상위 경로
    def sem_path(self):

        cur_sem = self.sem()[0]
        if cur_sem == None:
            return []
        filename = common_path+'/dict_semClassNum.txt'
        with open(filename, 'r',encoding="cp949") as file_object:
            lines = file_object.read()

        #print(lines)
        temp_list = []
        sem_list = []
        str = ""

        # 리스트 형성
        for tmp in lines:
            if tmp != '\n' and tmp != '\t':
                str += tmp
            else:
                if (str != ''):
                    sem_list.append(str)
                str = ''

        # 입력 단어 sem 위치 찾기
        regex = re.compile(r"_" + cur_sem + '$')
        for x in sem_list:
            if regex.search(x):
                cur_sem = x
                temp_list.append(cur_sem)

        while len(cur_sem.split('_')[0]) > 1:

            if cur_sem.split('_')[0][-2] == '.':
                tmp = cur_sem.split('_')[0][0:-2] + '_'
            else:
                tmp = cur_sem.split('_')[0][0:-3] + '_'
            regex = re.compile(r"^" + tmp)

            for x in sem_list:
                if regex.search(x):
                    cur_sem = x
                    temp_list.append(x)

        return list(reversed(temp_list))


    #유사도
    def wup_similarity(self,target):
        #self sem
        sem = self.sense.find("sem_grp")
        synn = sem.find("sem_class")
        synn1 = synn.text


        #target sem
        sem=target.sense.find("sem_grp")
        synn=sem.find("sem_class")
        synn2=synn.text


        list=[]
        path=common_path+"/layer.txt"
        f=open(path,'r')
        lines=f.readlines()
        for tmp in lines:
            if '_'+synn1+'\n' in tmp:
                list.append(tmp)
            if '_'+synn2+'\n' in tmp:
                list.append(tmp)

        ch=[]
        for tmp in list:
            ch.append(tmp.split("_")[0])

        word1 =ch[0].split('.');
        word2 =ch[1].split('.');

        same=0

        for tmp in range (0, min(len(word1),len(word2))):
            if word1[tmp] == word2[tmp]:
                same+=2
            else:
             break

        if self.name==target.name:
            same+=2

        result=same/((len(word1)+len(word2))+2)

        return result



# sense 바로 접근
def sense(input):

    input_list = input.split('.')
    arg= (input_list[0]+'.'+input_list[1]+'.'+input_list[2])
    target =entry(arg)
    allsense =target.entry.findall("sense")

    for se in allsense:
        if input==str(target.name+'.'+se.attrib['n']):
            return Sense(input,se,target.pos)

    #ss = str(self.name + "." + se.attrib['n'])
    #ss = str(self.name)

# entry 바로 접근
def entry(input):

    input_list = input.split('.')
    path=common_path+""
    if 'nn' in input_list[1]:
        path += "/01. 체언_상세"
    elif input_list[1] == 'vv':
        path += "/02. 용언_상세//vv"
    elif input_list[1] == 'va':
        path += "/02. 용언_상세//va"
    else:
        return


    path += "//"+input_list[0]+".xml"

    tree = parse(path)
    root = tree.getroot()
    allentry = root.findall("entry")
    for en in allentry:
        try:
            if input==str(input_list[0]+"."+en.attrib['pos']+"." + en.attrib['n']):
                return Entry(str(input_list[0]+"."+en.attrib['pos']+"." + en.attrib['n']), en, str(en.attrib['pos']))
        except KeyError:
            if input==str(input_list[0]+"."+en.attrib['pos']):
                return Entry(str(input_list[0]+"."+en.attrib['pos']), en, str(en.attrib['pos']))

# entry 객체 리턴
def entrys(word):
    path = filecheck(word)
    list = []

    for tmp in path:
        tree = parse(tmp)
        root = tree.getroot()
        allentry = root.findall("entry")

        for en in allentry:
            try:
                es = str(word + "." + en.attrib['pos'] + "." + en.attrib['n'])
            except KeyError:
                es = str(word + "." + en.attrib['pos'])

            temp = Entry(es, en, str(en.attrib['pos']))
            list.append(temp)

    return list


def _syn(word):

    ets=entrys(word)
    syn_list=[]

    for et in ets:
        for se in et.senses():
            syn_list+=se.syn()

    return syn_list

'''
def entry_error():

    path="./02. 용언_상세//va"
    abs_dir=os.path.join(os.getcwd(),path)
    file_names=os.listdir(abs_dir)

    #print(file_names)
    #print(len(file_names))
    error_list=[]


    for word in file_names:

        fpath=path+"//"+word
        #print(fpath)
        tree=parse(fpath)
        root=tree.getroot()
        #print(root.findtext('orth'))
        allentry=root.findall("entry")

        for en in allentry:
            try:
                en.attrib['n']

            except:

                error_list.append(word)
                break;

    print(error_list)
    print(len(error_list))
    print(len(file_names))


    return error_list



def sense_error():

    path="./02. 용언_상세//va"
    abs_dir=os.path.join(os.getcwd(),path)
    file_names=os.listdir(abs_dir)

    error_list=[]


    for word in file_names:

        fpath=path+"//"+word
        tree=parse(fpath)
        root=tree.getroot()
        allentry=root.findall("entry")

        for en in allentry:
            allsense=en.findall("sense")
            for se in allsense:
                try:
                    se.attrib['n']
                except:

                    if word not in error_list:
                        error_list.append(word)
                    break;

    print(error_list)
    print(len(error_list))
    print(len(file_names))


    return error_list

'''

# file check
def filecheck(word):
    n_path = common_path+"/01. 체언_상세"
    vv_path = common_path+"/02. 용언_상세/vv"
    va_path = common_path+"/02. 용언_상세/va"


    path = [n_path, vv_path, va_path]
    ret_list = []
    check = word + ".xml"


    for tmp in path:
        if check in os.listdir(tmp):
            ret_list.append(tmp + "/" + check)

    return ret_list



# -*- coding: utf-8 -*-

"""
Class for dealing with SRL data.
"""

from collections import defaultdict
import _pickle as cPickle
import _pickle
import logging
import re
import os
import numpy as np
#from itertools import izip

from .. import attributes
from .. import utils
from ..word_dictionary import WordDictionary
from .. import reader

class ConllPos(object):
	"""
	Dummy class for storing the position of each field in a
	CoNLL data file.
	"""
	id = 0
	word = 1
	lemma = 2
	hmorph = 3		# head morph
	hpos = 4			# head pos
	tmorph = 5		# tail morph
	tpos = 6			# tail pos
	parse = 7
	rel = 8
	semantic_role = 9
	SEP = '\t'

class SRLReader(reader.TaggerReader):
	
	def __init__(self, md=None, filename=None):
		"""
		The reader will read sentences from a given file. This file must
		be in the correct format (one token per line, columns indicating
		which tokens are predicates and their argument structure).
		
		:param filename: a file with CoNLL-like format data. If it is None,
			the reader will be created with no data.
		:param only_boundaries: train to identify only argument boundaries
		:param only_classify: train to classify pre-determined argument
		:param only_predicates: train to identify only predicates
		"""
		
		self.taskname = 'srl'
		self.pos_dict = {}
		
		if filename is not None:
			self._read_conll(filename)
			#self._clean_text()
		
		super(SRLReader, self).__init__(md)
		
	
	@property
	def task(self):
		"""
		Abstract Base Class (ABC) attribute.
		"""
		return self.taskname
	
	
	def _read_conll(self, filename):
		'''
		Read a file in CoNLL format and extracts semantic role tags
		for each token.
		'''
		lines = []
		with open(filename, 'rt') as f:
			for line in f:
				line = line.strip()
				lines.append(line)
		
		self.sentences = []
		self.predicates = []
		tokens = []
		sent_predicates = []
		sent_tags = []
		token_number = 0
		
		for line in lines:
			line = line.strip()
			
			if line == '':
				# blank line between sentences
				if len(tokens) > 0:
					sentence = (tokens, sent_tags)
					self.sentences.append(sentence)
					self.predicates.append(np.array(sent_predicates))
					tokens = []
					sent_predicates = []
					sent_tags = []
					token_number = 0
				
				continue
			
			fields = line.split(ConllPos.SEP)
			idx = fields[ConllPos.id]
			word = fields[ConllPos.word]
			lemma = fields[ConllPos.lemma]
			hmorph = fields[ConllPos.hmorph]
			hpos = fields[ConllPos.hpos].lower()
			tmorph = fields[ConllPos.tmorph]
			tpos = fields[ConllPos.tpos].lower()
			parse = fields[ConllPos.parse]
			rel = fields[ConllPos.rel]
			is_predicate = (rel[:1] == 'V')
			tag = fields[ConllPos.semantic_role]
			
			tag = self._read_role(tag)	
			sent_tags.append((int(parse)-1, tag))			# note: codify_sentences

			token = attributes.Token(word=word, morph_h=hmorph, morph_t=tmorph, pos_t=tpos, chunk=rel)
			#token = attributes.Token(word, morph_h=hmorph, pos_h=hpos, morph_t=tmorph, pos_t=tpos, chunk=rel)
			tokens.append(token)
			if is_predicate:
				sent_predicates.append(token_number)
			
			token_number += 1
		
		if len(tokens) > 0:
			# last sentence
			sentence = (tokens, sent_tags)
			self.sentences.append(sentence)
			self.predicates.append(np.array(sent_predicates))
	
	@classmethod
	def _read_role(cls, role):
		'''
		Reads the semantic role from a CoNLL-style file.

		pram: role what is read from the conll file
		'''
		return role 

	def extend(self, data):
		"""
		Adds more data to the reader.
		:param data: a list of tuples in the format (tokens, tags, predicates), 
		one for each sentence.
		"""
		self.sentences.extend([(sent, tags) for sent, tags, _ in data])
		self.predicates.extend([np.array(preds) for _, _, preds in data])
	
	def load_or_create_tag_dict(self):
		"""
		In the case of SRL argument classification or one step SRL, try to 
		load the tag dictionary. If the file with the tags is not present,
		a new one is created from the available sentences. 
		
		In the case of argument detection or predicate detection, 
		this function does nothing.
		"""
		if os.path.isfile(self.md.paths['srl_tags']):
			self.load_tag_dict()
			return
		
		self._create_tag_dict()
		logger = logging.getLogger('Logger')
		logger.info('Created SRL tag dictionary')
	
	def _create_tag_dict(self):
		"""
		Examine the available sentences and create a tag dictionary.
		
		:param iob: If True, this function will generate an entry for B-[tag] 
			and one for I-[tag], except for the tag 'O'.
		"""
		logger = logging.getLogger("Logger")
		tags = {tag
				for _, tags in self.sentences
				for rel, tag in tags}
		
		# create a dictionary now even if uses IOB, in order to save it in 
		# a deterministic order
		self.tag_dict = {tag: code for code, tag in enumerate(tags)}
		reader.save_tag_dict(self.md.paths['srl_tags'], self.tag_dict)
		logger.debug("Saved SRL tag dictionary.")
		
	
	def load_tag_dict(self, filename=None, iob=False):
		"""
		Loads the tag dictionary from the default file. The dictionary file should
		have one tag per line. 
		
		:param iob: If True, this function will generate an entry for B-[tag] 
			and one for I-[tag], except for the tag 'O'.
		"""
		if filename is None:
			filename = self.md.paths['srl_tags']
		
		self.tag_dict = {}
		code = 0
		with open(filename, 'rt') as f:
			for tag in f:
				tag = tag.strip()
				if tag == '':
					continue
				
				self.tag_dict[tag] = code
				
				code += 1
		
	
	def _generate_iobes_dictionary(self):
		"""
		Generate the reader's tag dictionary mapping the IOBES tags to numeric codes.
		"""
		self.tag_dict = {tag: code for code, tag in enumerate('IOBES')}
	
	def _generate_predicate_id_dictionary(self):
		"""
		Generate a tag dictionary for identifying predicates.
		It has two tags: V for predicates and O for others.
		"""
		self.tag_dict = {'-': 0, 'V': 1}
		#self.tag_dict = {'O': 0, 'V': 1}
	
	def generate_dictionary(self, dict_size=None, minimum_occurrences=2):
		"""
		Generates a token dictionary based on the given sentences.
		
		:param dict_size: Max number of tokens to be included in the dictionary.
		:param minimum_occurrences: Minimum number of times that a token must
			appear in the text in order to be included in the dictionary.
		"""
		logger = logging.getLogger("Logger")
		all_tokens = [token.word
					  for tokens, _ in self.sentences
					  for token in tokens]
		self.word_dict = WordDictionary(all_tokens, dict_size, minimum_occurrences)
		logger.info("Created dictionary with %d tokens" % self.word_dict.num_tokens)
	
	def _clean_text(self):
		"""
		Cleans the sentences text, replacing numbers for a keyword, different
		kinds of quotation marks for a single one, etc.
		"""
		for sent, _ in self.sentences:
			for i, token in enumerate(sent):
				new_word = utils.clean_text(token.word, correct=False)
				new_lemma = utils.clean_text(token.lemma, correct=False) 
				token.word = new_word
				token.lemma = new_lemma
				sent[i] = token

	def create_converter(self):
		"""
		This function overrides the TextReader's one in order to deal with Token
		objects instead of raw strings.
		"""
		self.converter = attributes.TokenConverter()
		
		if self.md.use_lemma:
			# look up word lemmas 
			word_lookup = lambda t: self.word_dict.get(t.lemma)
		else:
			# look up the word itself
			word_lookup = lambda t: self.word_dict.get(t.word)
			 
		self.converter.add_extractor(word_lookup)
		
		#if self.md.use_caps:
		#	caps_lookup = lambda t: attributes.get_capitalization(t.word)
		#	self.converter.add_extractor(caps_lookup)
		
		if self.md.use_pos:
			with open(self.md.paths['pos_tag_dict']) as f:
				#pos_dict = cPickle.load(f)
				buf = f.readlines()
				for i, line in enumerate(buf):
					line = line.strip()
					self.pos_dict[line] = i

			pos_def_dict = defaultdict(lambda: self.pos_dict['NN'])
			pos_def_dict.update(self.pos_dict)
			pos_lookup = lambda t: pos_def_dict[t.pos_t]
			self.converter.add_extractor(pos_lookup)
		
		#if self.md.use_chunk:
		#	with open(self.md.paths['chunk_tag_dict']) as f:
		#		chunk_dict = cPickle.load(f)
			
		#	chunk_def_dict = defaultdict(lambda: chunk_dict['O'])
		#	chunk_def_dict.update(chunk_dict)
		#	chunk_lookup = lambda t: chunk_def_dict[t.chunk]
		#	self.converter.add_extractor(chunk_lookup)
	
	def get_num_pos_tags(self):
		return len(self.pos_dict)

	def generate_tag_dict(self):
		"""
		Generates a tag dictionary that converts the tag itself
		to an index to be used in the neural network.
		"""
		self.tagset = set(tag
						  for _, props in self.sentences
						  for prop in props
						  for _, tag in prop)
		
		self.tag_dict = dict( zip( self.tagset,
								   xrange(len(self.tagset))
								   )
							 )
	
	def _remove_tag_names(self):
		"""Removes the actual tag names, leaving only IOB or IOBES block delimiters."""
		for _, propositions in self.sentences:
			for tags in propositions:
				for i, (_, tag) in enumerate(tags):
					tags[i] = tag[0]
	
	def _codify_sentences(self):
		"""Internal helper function."""
		new_sentences = []
		self.tags = []
		
		for (sent, props) in self.sentences:
			new_sent = []
			sentence_tags = []
			
			for token in sent:
				new_token = self.converter.convert(token)
				new_sent.append(new_token)
			
			for prop in props:
				sentence_tags.append(prop)

			new_sentences.append(np.array(new_sent))
			self.tags.append(sentence_tags)
		#print(new_sentences, flush=True)
		#print(self.tags, flush=True)
		
		self.sentences = new_sentences
		self.codified = True
	
	def codify_sentences(self):
		"""
		Converts each token in each sequence into indices to their feature vectors
		in feature matrices. The previous sentences as text are not accessible anymore.
		Tags are also encoded. This function takes care of the case of classifying 
		pre-delimited arguments.
		"""
		if self.converter is None:
			self.create_converter()
		
		self._codify_sentences()
		self.arg_limits = []
		
		for i, propositions in enumerate(self.tags):
			new_sent_tags = []
			sent_args = []
			
			for j, (rel, prop_tags) in enumerate(propositions):
				
				new_prop_tags = []
				prop_args = []
				
				#if prop_tags != '-' and j==rel:
				if prop_tags != '-' :
					prop_args.append(np.array([j, j+1]))
					new_prop_tags.append(self.tag_dict[prop_tags])

				sent_args.append(np.array(prop_args))
				new_sent_tags.append(np.array(new_prop_tags))
			
			self.arg_limits.append(sent_args)
			self.tags[i] = new_sent_tags
					 
	
	def convert_tags(self, scheme, update_tag_dict=True, only_boundaries=False):
		"""
		Replaces each word label with an IOB or IOBES version, appending a prefix
		to them. 
		
		:param scheme: IOB or IOBES (In, Other, Begin, End, Single).
		:param update_dict: whether to update or not the tag dictionary after
			converting the tags.
		:param only_boundaries: if True, only leaves the IOBES tags and remove
			the actual tags. Also, avoid updating the tag dict.
		"""
		scheme = scheme.lower()
		if scheme not in ('iob', 'iobes'):
			raise ValueError("Unknown tagging scheme: %s" % scheme)
		
		for _, props in self.sentences:
			for prop in props:
				
				last_tag = None
				for i, tag in enumerate(prop):
					
					if tag == 'O':
						# O tag is independent from IBES
						last_tag = tag 
						continue
					
					try:
						next_tag = prop[i + 1]
					except IndexError:
						# last word already
						next_tag = None
					 
					if tag != last_tag:
						# a new block starts here. 
						last_tag = tag
						if scheme == 'iob' or next_tag == tag:
							prop[i] = 'B-%s' % tag
						else:
							prop[i] = 'S-%s' % tag
					else:
						# the block continues. 
						if scheme == 'iob' or next_tag == tag:
							prop[i] = 'I-%s' % tag
						else:
							prop[i] = 'E-%s' % tag
			
		if only_boundaries:
			self._remove_tag_names()
		elif update_tag_dict:
			self.generate_tag_dict()
		else:
			# treat any tag not appearing in the tag dictionary as O
			actual_tagset = {tag for _, props in self.sentences for prop in props for tag in prop}
			for tag in actual_tagset:
				if tag not in self.tag_dict:
					self.tag_dict[tag] = self.tag_dict[self.rare_tag]
	



"""
string2string similarity
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""


"""
    This class contains the original implementation of the BARTScore algorithm by Yuan et al. (2021).

    BARTScore: BART-based Evaluation Metric for Text Generation

    @inproceedings{bartscore2021,
        author = {Yuan, Weizhe and Neubig, Graham and Liu, Pengfei},
        booktitle = {Advances in Neural Information Processing Systems},
        editor = {M. Ranzato and A. Beygelzimer and Y. Dauphin and P.S. Liang and J. Wortman Vaughan},
        pages = {27263--27277},
        publisher = {Curran Associates, Inc.},
        title = {BARTScore: Evaluating Generated Text as Text Generation},
        url = {https://proceedings.neurips.cc/paper/2021/file/e4d2b6e6fdeca3e60e0f1a62fee3d9dd-Paper.pdf},
        volume = {34},
        year = {2021}
    }

    Disclaimer:
        This code is adapted from https://github.com/neulab/BARTScore/blob/main/bart_score.py
"""

import numpy as np
from typing import List, Union, Dict
import traceback
from nltkor.make_requirement import make_requirement

import torch
import torch.nn as nn
from transformers import BartTokenizer, BartForConditionalGeneration


# BARTScore class
class BARTScore:
    """
    This class implements the BARTScore algorithm.
    """

    def __init__(self,
        model_name_or_path='facebook/bart-large-cnn',
        tokenizer_name_or_path: str = None,
        device: str = 'cpu',
        max_length=1024,
        ) -> None:
        r"""
        This function initializes the BARTScore class, which computes the BARTScore between two pieces of text.

        Arguments:
            model_name_or_path (str): The name or path of the model. Defaults to 'facebook/bart-large-cnn'.
            tokenizer_name_or_path (str): The name or path of the tokenizer. Defaults to None.
            device (str): The device to use. Defaults to 'cpu'.
            max_length (int): The maximum length of the input. Defaults to 1024.

        Returns:
            None

        Raises:
            ValueError: If the device is not 'cpu' or 'cuda'.

         .. attention::

            If you use this class, please make sure to cite the following paper:

            .. code-block:: latex

                @inproceedings{bartscore2021,
                    author = {Yuan, Weizhe and Neubig, Graham and Liu, Pengfei},
                    booktitle = {Advances in Neural Information Processing Systems},
                    editor = {M. Ranzato and A. Beygelzimer and Y. Dauphin and P.S. Liang and J. Wortman Vaughan},
                    pages = {27263--27277},
                    publisher = {Curran Associates, Inc.},
                    title = {BARTScore: Evaluating Generated Text as Text Generation},
                    url = {https://proceedings.neurips.cc/paper/2021/file/e4d2b6e6fdeca3e60e0f1a62fee3d9dd-Paper.pdf},
                    volume = {34},
                    year = {2021}
                }

        .. note::
            * The default model is the BART-large-cnn model.
            * If the tokenizer name or path is not specified, then the model name or path will be used.
            * If the device is 'cuda', then the model will be loaded onto the GPU.
            * If device is not specified, use the GPU if available, otherwise use the CPU.

        """

        if tokenizer_name_or_path is None:
            tokenizer_name_or_path = model_name_or_path

        # Set the attributes
        self.device = device
        self.max_length = max_length

        # Load model and tokenizer
        self.tokenizer = BartTokenizer.from_pretrained(tokenizer_name_or_path)
        self.model = BartForConditionalGeneration.from_pretrained(model_name_or_path)
        self.model.eval()
        self.model.to(device)

        # Set up loss
        self.loss_fct = nn.NLLLoss(reduction='none', ignore_index=self.model.config.pad_token_id)
        self.lsm = nn.LogSoftmax(dim=1)



    # Loads the model weights from a specified path
    def load(self,
        weights_path=None,
        ) -> None:
        """
        This function loads the model weights from a specified path.

        Arguments:
            weights_path (str): The path to the weights.

        Returns:
            None
        """
        if weights_path is None:
            weights_path = 'models/bart.pth'

        self.model.load_state_dict(torch.load(weights_path, map_location=self.device))



    # Compute the BARTScore between source sentences and target sentences
    def compute(self,
        source_sentences: List[str],
        target_sentences: Union[List[str], List[List[str]]],
        batch_size: int = 4,
        agg: str = 'mean',
        ) -> Dict[str, List[float]]:
        """
        This function scores the target sentences against the source sentences using BARTScore.

        Arguments:
            source_sentences (List[str]): The source sentences.
            target_sentences (Union[List[str], List[List[str]]]): The target sentences.
            batch_size (int): The batch size to use (default: 4)
            agg (str): The aggregation method. Defaults to 'mean'; used only when target_sentences is a list of lists.

        Returns:
            Dict[str, List[float]]: The BARTScore for each example.

        Raises:
            ValueError: If the number of source sentences and target sentences do not match.
        """
        # Check the number of source sentences and target sentences
        if len(source_sentences) != len(target_sentences):
            raise ValueError(f'Number of source sentences ({len(source_sentences)}) and number of target sentences ({len(target_sentences)}) do not match.')

        # If the target sentences are a list of lists, then call the multi_ref_score function
        if isinstance(target_sentences[0], list):
            return self.compute_multi_ref_score(
                source_sentences=source_sentences,
                target_sentences=target_sentences,
                batch_size=batch_size,
                agg=agg
            )

        # Score for each example
        score_list = []

        for i in range(0, len(source_sentences), batch_size):
            # Get the current batch
            src_batch = source_sentences[i: i + batch_size]
            tgt_batch = target_sentences[i: i + batch_size]
            try:
                with torch.no_grad():
                    # Encode the batch
                    encoded_src = self.tokenizer(
                        src_batch,
                        max_length=self.max_length,
                        truncation=True,
                        padding=True,
                        return_tensors='pt'
                    )
                    encoded_tgt = self.tokenizer(
                        tgt_batch,
                        max_length=self.max_length,
                        truncation=True,
                        padding=True,
                        return_tensors='pt'
                    )

                    # Get the input ids and attention masks for the source and target sentences
                    src_tokens = encoded_src['input_ids'].to(self.device)
                    src_mask = encoded_src['attention_mask'].to(self.device)
                    tgt_tokens = encoded_tgt['input_ids'].to(self.device)
                    tgt_mask = encoded_tgt['attention_mask']
                    tgt_len = tgt_mask.sum(dim=1).to(self.device)

                    # Feed the batch to the model and get the loss
                    output = self.model(
                        input_ids=src_tokens,
                        attention_mask=src_mask,
                        labels=tgt_tokens
                    )
                    logits = output.logits.view(-1, self.model.config.vocab_size)
                    # Compute the loss
                    loss = self.loss_fct(self.lsm(logits), tgt_tokens.view(-1))
                    loss = loss.view(tgt_tokens.shape[0], -1)
                    loss = loss.sum(dim=1) / tgt_len
                    # Get the score
                    curr_score_list = [-x.item() for x in loss]
                    # Append the score to the list
                    score_list += curr_score_list

            except:
                # If there is an error, print the traceback
                raise Exception(f'Error in scoring batch {i // batch_size}:\n{traceback.format_exc()}')
        return {'score': np.array(score_list)}



    # Score a batch of examples with multiple references
    def compute_multi_ref_score(self,
        source_sentences: List[str],
        target_sentences: List[List[str]],
        batch_size: int = 4,
        agg: str = "mean",
        ) -> Dict[str, List[float]]:
        """
        Score a batch of examples with multiple references.

        Arguments:
            source_sentences (List[str]): The source sentences.
            target_sentences (List[List[str]]): The target sentences.
            agg (str): The aggregation method. Can be "mean" or "max".
            batch_size (int): The batch size.

        Returns:
            Dict[str, List[float]]: The BARTScore for each example.

        Raises:
            ValueError: If the number of source sentences and target sentences do not match.
        """

        # Assert we have the same number of references
        ref_nums = [len(x) for x in target_sentences]
        if len(set(ref_nums)) > 1:
            raise Exception("You have different number of references per test sample.")

        ref_num = len(target_sentences[0])
        score_matrix = []
        for i in range(ref_num):
            curr_target_sentences = [x[i] for x in target_sentences]
            scores = self.compute(source_sentences, curr_target_sentences, batch_size)
            score_matrix.append(scores)
        if agg == "mean":
            score_list = np.mean(score_matrix, axis=0)
        elif agg == "max":
            score_list = np.max(score_matrix, axis=0)
        else:
            raise NotImplementedError(f"Aggregation method {agg} not implemented yet.")
        return {"score": score_list}

def demo():
    demo_setences = [
        ("I am a student", "He is a teacher"),
        ("나는 학생이다", "그는 선생님이다"),
        ("점심에 온기동에서 삼겹차슈덮밥을 먹었다.", "저녁에 피나치공에서 피자와 치킨을 먹었다."),
        ('제가 나와 있는 곳은 경남 거제시 옥포동 덕포 해수욕장에 나와 있습니다.', '강한 바람에 간판이나 지붕이 떨어지는 등 피해가 잇따르기도 했습니다.'),
        ('Outraged mortuary workers in Kenya have criticised the country’s police chief after he accused them of leasing corpses to opposition politicians.',
        'Head of police Japheth Koome earlier this week claimed that opposition politicians hired bodies from mortuaries and planted them at the scenes of protests so as to blame the police for brutality.')

    ]
    for str1, str2 in demo_setences:
        print("demo : ", BARTScore().compute([str1], [str2]))

if __name__ == "__main__":
    demo()



import os
import sys
import argparse
import numpy as np
from typing import Callable, Iterable, List, Tuple, Union
from copy import deepcopy
import itertools
import torch
import time
import math
from nltk.translate.bleu_score import *
from nltk.metrics import confusionmatrix
from collections import defaultdict
from nltk.util import ngrams, skipgrams
#from nltkor.tokenize.ko_tokenize import word_tokenize,sent_tokenize,syllable_tokenize
from nltkor.make_requirement import make_requirement
from nltkor.tokenize import Ko_tokenize
from nltkor.cider.cider import Cider
from nltkor.tag import EspressoTagger
from nltkor.sejong import ssem

try:
    import torch
except ImportError:
	file_path = make_requirement(['torch'])
	raise Exception(f"""
    Need to install Libraries, please pip install below libraries
    \t pip install torch
	Or, use pip install requirement.txt
    \t  pip install -r {file_path}
    """)


class DefaultMetric:

	def __init__(self, lang="ko"):
		self.lang = lang
		#if lang not in ["korean", "english"]:
		if lang not in ["ko", "en"]:
			raise Exception("Only \"korean\" or \"english\" in lang parameter")
		self.tokenize=lambda ref: Ko_tokenize.word(ref,self.lang)
		self.skip_pos = ['EE']

	def accuracy_score(self, true, pred):

		mat=confusionmatrix.ConfusionMatrix(true,pred)

		conf=mat._confusion
		total=0
		tp=0

		for r, tmp in enumerate(conf):
			for v, n in enumerate(tmp):
				if r==v:
					tp+=n
				total+=n

		return float(tp/total)

	def accuracy_norm(model, tokenizer, input_text: str, candidates: list, label: int):
		reserved_memory = []
		inference_time = []
		tokenized_prompt = tokenizer(input_text, return_tensors='pt').input_ids
		total_candidate = []

		for ending in candidates:
			len_ending = len(ending)
			tokenized_ending = tokenizer(ending, return_tensors='pt').input_ids
			tokenized_ending = tokenized_ending[:, 1:]
			input_ids = torch.cat([tokenized_prompt, tokenized_ending], dim=-1).cuda()
			labels = input_ids.clone()
			labels[0, :tokenized_prompt.shape[1]] = -100
			start = time.time()
			with torch.no_grad():
				outputs = model(input_ids, labels=labels)
				inference_time.append(time.time() - start)
			reserved_memory.append(torch.cuda.memory_reserved() / (1024**2))
			total_logprobs = -outputs.loss.item() * tokenized_ending.shape[1]
			total_candidate.append(total_logprobs/len_ending)
		answer_idx = total_candidate.index(max(total_candidate))
		if int(label) == answer_idx:
			cor = 1
		else:
			cor = 0
		metric_dict = {
			"reserved_memory": reserved_memory,
			"inference_time": inference_time
			}
		return cor, metric_dict

	def recall_score(self, true, pred, avg='micro'):

		mat=confusionmatrix.ConfusionMatrix(true,pred)

		conf=mat._confusion
		indices=mat._indices
		values=mat._values
		total=0


		if len(values)==2:
			tp=0
			fn=0
			for r,i in enumerate(conf):
				for r2,v in enumerate(i):
					if r==0:
						continue
					elif r2==0:
						fn=v
					elif r==1:
						tp=v

			return float(tp/(tp+fn))


		c_tp=[]
		c_fn=[]
		recall_cls=[]

		for r, tmp in enumerate(conf):
			temp=0
			for v, n in enumerate(tmp):
				if r==v:
					c_tp.append(n)
				else:
					temp+=n
			c_fn.append(temp)

		if avg=='macro':

			for tmp in range(len(values)):
				try:
					recall_cls.append(float(c_tp[tmp]/(c_tp[tmp]+c_fn[tmp])))
				except:
					recall_cls.append(0)

			temp=0

			for tmp in recall_cls:
				temp+=tmp

			return float(temp/len(recall_cls))

		if avg=='micro':
			ja=0
			mo=0

			for tmp in range(len(values)):
				ja+=c_tp[tmp]
				mo+=c_tp[tmp]+c_fn[tmp]

			return float(ja/mo)

		else:
			return "avg expect micro/macro"



	def precision_score(self, true, pred,avg='micro'):


		mat=confusionmatrix.ConfusionMatrix(true,pred)

		conf=mat._confusion
		values=mat._values

		total=0

		if len(values)==2:
			tp=0
			fp=0
			for r,i in enumerate(conf):
				for r2,v in enumerate(i):
					if r2==0:
						continue
					elif r==0:
						fp=v
					elif r==1:
						tp=v

			return float(tp/(tp+fp))

		c_tp=list()
		c_fp=[0 for _ in range(len(values))]
		recall_cls=[]

		for r, tmp in enumerate(conf):
			for v, n in enumerate(tmp):
				if r==v:#tp
					c_tp.append(n)
				else:
					c_fp[v]+=n

		if avg=='macro':
			for tmp in range(len(values)):
				try:
					recall_cls.append(float(c_tp[tmp]/(c_tp[tmp]+c_fp[tmp])))
				except:
					recall_cls.append(0)

			temp=0

			for tmp in recall_cls:
				temp+=tmp

			return float(temp/len(recall_cls))


		elif avg=='micro':
			ja=0
			mo=0

			for tmp in range(len(values)):
				ja+=c_tp[tmp]
				mo+=c_tp[tmp]+c_fp[tmp]

			return float(ja/mo)

		else:
			return "avg expect micro/macro"


	def f1_score(self, true, pred, avg='micro'):

		if avg =='micro' or avg =='macro':

			precision=self.precision_score(true,pred,avg)
			recall=self.recall_score(true,pred,avg)
		else:
			return "avg expect micro/macro"

		return (((precision*recall)/(precision+recall))*2)

    


	def pos_eval(self, fin):

		#temp=os.getcwd()+'/'+fin
		file=open(fin,'r').read()
		sents=file.split("\n\n")

		acc = defaultdict(float)
		t_avg = defaultdict(float)

		for sent in sents:
			lines=sent.split('\n')
			for line in lines:
				tot=line.split('\t')

				if line=='':continue

				wd=tot[0]
				gold=tot[1]
				pred=tot[2]

				acc['all']+=1
				gold_list=gold.split('+')
				pred_list=pred.split('+')

				t_avg["pr_all"]+=len(pred_list)
				t_avg["rc_all"]+=len(gold_list)

				if gold==pred:
					acc["true"]+=1
					t_avg['pr']+=len(pred_list)
					t_avg['rc']+=len(gold_list)
					continue
				else :
					intersect=0
					for g in gold_list:
						if not g in pred_list: continue
						intersect+=1
					t_avg['pr']+=intersect
					t_avg['rc']+=intersect


		t_avg['pr_result'] = t_avg['pr'] / t_avg['pr_all']
		t_avg['rc_result'] = t_avg['rc'] / t_avg['rc_all']

		return float(acc['true']/acc['all']) ,t_avg['pr_result'],t_avg['rc_result'], f1(t_avg['pr_result'], t_avg['rc_result'])


	def f1(self, p, r):
		return 2 * p * r / (p + r) if p + r else 0


	def precision_at_k(self, true: List[int], pred: List[int], k: int) -> float:
		"""
		avg = ['micro', 'macro']
		"""

		relevant = 0

		if k > len(pred):
			raise ValueError("`k` is bigger than pred's length")

		pred = pred[:k]

		for t in true:
			if t in pred:
				relevant += 1


		return float(relevant/len(pred))

	def recall_at_k(self, true: List[int], pred: List[int], k: int) -> float:

		relevant = 0

		if k > len(pred):
			raise ValueError("`k` is bigger than pred's length")

		pred = pred[:k]

		for t in true:
			if t in pred:
				relevant += 1


		return float(relevant/len(true))

	def hit_rate_at_k(self, user: List[List[int]], pred: List[List[int]], k: int) -> float:
		hit = 0

		for u_list, p_list in zip(user, pred):
			try:
				p_list = p_list[:k]
			except:
				raise ValueError("`k` is bigger than pred's length ")
			for u in u_list:
				if u in p_list:
					hit += 1
					break

		return float(hit/len(user))

	def mean_absolute_error(self, true: Union[torch.Tensor, np.ndarray], pred: Union[torch.Tensor, np.ndarray]) -> float:
		pass

	def root_mean_square_error(self, true: Union[torch.Tensor, np.ndarray], pred: Union[torch.Tensor, np.ndarray]) -> float:
		pass

	def _W_CER(self, r, h):

		costs = [[0 for inner in range(len(h)+1)] for outer in range(len(r)+1)]

		DEL_PENALTY=1 # Tact
		INS_PENALTY=1 # Tact
		SUB_PENALTY=1 # Tact

		for i in range(1, len(r)+1):
			costs[i][0] = DEL_PENALTY*i

		for j in range(1, len(h) + 1):
			costs[0][j] = INS_PENALTY*j

		# computation
		for i in range(1, len(r)+1):
			for j in range(1, len(h)+1):
				if r[i-1] == h[j-1]:
					costs[i][j] = costs[i-1][j-1]
				else:
					substitutionCost = costs[i-1][j-1] + SUB_PENALTY # penalty is always 1
					insertionCost	= costs[i][j-1] + INS_PENALTY   # penalty is always 1
					deletionCost	 = costs[i-1][j] + DEL_PENALTY   # penalty is always 1

					costs[i][j] = min(substitutionCost, insertionCost, deletionCost)

		mo = len(r)
		i = len(r)
		j = len(h)

		result=(costs[i][j])/mo

		if result>1.0:
			return 1.0
		else:
			return result


	def wer(self, reference, candidate):
		r = Ko_tokenize.word(reference)
		h = Ko_tokenize.word(candidate)

		return self._W_CER(r,h)


	def cer(self, reference,candidate):
		r = Ko_tokenize.syllable(reference)
		h = Ko_tokenize.syllable(candidate)

		return self._W_CER(r,h)


	def bleu(self, reference, candidate,weights=(0.25,0.25,0.25,0.25), smoothing_function=None):

		if type(candidate)!=list or type(reference)!=list:
			print("parameter expect list type")
			return

		reference=list(map(self.tokenize,reference))
		candidate=Ko_tokenize.word(candidate)

		return sentence_bleu(reference,candidate,weights,smoothing_function=smoothing_function)


	def bleu_n(self, reference,candiate,n=1, smoothing_function=None):

		if n==1:
			return self.bleu(reference,candiate,(1,0,0,0), smoothing_function=smoothing_function)
		elif n==2:
			return self.bleu(reference,candiate,(0,1,0,0), smoothing_function=smoothing_function)
		elif n==3:
			return self.bleu(reference,candiate,(0,0,1,0), smoothing_function=smoothing_function)
		elif n==4:
			return self.bleu(reference,candiate,(0,0,0,1), smoothing_function=smoothing_function)

	def bleu_tensor(self,reference,candidate,n=0, smoothing_function=None):

		if n: weights = tuple(1 if i == n-1 else 0 for i in range(4))
		else: weights = (0.25, 0.25, 0.25, 0.25)
 
		reference=reference.unsqueeze(1)
		reference=reference.numpy()
		candidate=candidate.numpy()
		return torch.tensor(corpus_bleu(reference,candidate,weights,smoothing_function=smoothing_function))





	def _hyp_sent_split_remove(self, can):

		can_sent=[[tmp.rstrip('.'or'?'or'!'or','or'\n')] for tmp in Ko_tokenize.sentence(can)]
		return can_sent

	def _ref_sent_split_remove(self, ref):

		ref_sent=[Ko_tokenize.sentence(tmp) for tmp in ref]
		ref_sent_c=[]
		for tmp in ref_sent:
			ref_sent_in=[]
			for tmp2 in tmp:
				ref_sent_in.append(Ko_tokenize.word(tmp2.rstrip('.'or'?'or'!'or','or'\n')))
			ref_sent_c.append(ref_sent_in)

		return ref_sent_c

	def _token(self, ref_stoken, can, n):

		numer=[]
		ref_len=0

		can=list(ngrams(can,n))

		for tmp in ref_stoken:

			if n==1:
				ref=list(ngrams(tmp,1))
			elif n==2:
				ref=list(ngrams(tmp,2))
			elif n==3:
				ref=list(ngrams(tmp,3))
			else:return 0

			intersect = [t for t in ref if t in can ]
			numer.append(len(intersect))
			ref_len+=len(ref)

		try:
			rec= sum(numer)/ref_len
		except:
			rec=0

		return rec



	def rouge_n(self, ref, can, n=1):

		beta=1
		rec,prec=0,0

		can_sent=self._hyp_sent_split_remove(can)
		can_word=list(itertools.chain(*[Ko_tokenize.word(tmp,self.lang) for tmp in can_sent]))
		ref=self._ref_sent_split_remove(ref)

		r_list=[]

		for tmp in ref:
			if n==1:
				r_list.append(self._token(tmp, can_word, 1))
			elif n==2:
				r_list.append(self._token(tmp, can_word, 2))
			elif n==3:
				r_list.append(self._token(tmp, can_word, 3))

		return max(r_list)



	def rouge_l(self, ref, can):

		beta=1
		#check=0

		can= self._hyp_sent_split_remove(can)
		can=[Ko_tokenize.word(tmp,self.lang) for tmp in can]
		refs=self._ref_sent_split_remove(ref)

		can_word=list(itertools.chain(*can))

		result_list=[]

		for ref in refs:
			lcs_list=[]
			for ri in ref:
				ri_C=[]
				for ci in can:
					temp=self._lcs(ci,ri)
					ri_C.append(temp)

				ri_C=list(itertools.chain(*ri_C))
				ri_C=set(ri_C)
				lcs_list.append(len(ri_C))

			ref_word=list(itertools.chain(*ref))

			R_lcs=sum(lcs_list)/len(ref_word)
			P_lcs=sum(lcs_list)/len(can_word)

			try:
				F_lcs= (2*R_lcs*P_lcs)/(R_lcs+P_lcs)
			except:
				F_lcs=0
			result_list.append(F_lcs)

		return max(result_list)



	def _lcs(self, can, ref):


		s1=can
		s2=ref
		check=0

		if len(s1)<=len(s2):
			temp=s1
			s1=s2
			s2=temp
			check=1

		m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]

		for x in range(1, 1 + len(s1)):
			for y in range(1, 1 + len(s2)):
				if s1[x - 1] == s2[y - 1]:
					m[x][y] = m[x - 1][y - 1] +1
				else:
					m[x][y]=max(m[x][y-1],m[x-1][y])
		f_x=len(s2)+1
		lcs=m[len(s1)][len(s2)]
		temp=[]


		i=len(s1)
		j=len(s2)

		while m[i][j]!=0:
			if(m[i][j]==m[i][j-1]):
				j-=1
			elif (m[i][j]==m[i-1][j]):
				i-=1
			else:
				if check==0:
					temp.append(s1[i-1])
				if check==1:
					temp.append(s2[j-1])
				i-=1
				j-=1

		return temp
		'''
		for y in reversed(range(1,1+len(s1))):
			for x in reversed(range(1,1+len(s2))):
				if (m[y][x]-m[y-1][x-1]==1) and (m[y][x]-m[y-1][x]==1) and (m[y][x]-m[y][x-1]==1):
					if (y==len(s1)+1) and (x==len(s2)):
						temp.append(x)
					else:
						temp.append(x-1)

		print('the police 만  나와줘야',temp)
		if check==0:
			word=s1
		elif check==1:
			word=s2

		ret_list=[]

		for tmp in range(len(temp)):
			ret_list.append(word[temp[tmp]])

		return ret_list
		'''


	def _skip_bigrams(self, ref_stoken, can_sent, can, n=1):

		beta=1
		numer=[]
		ref_len=0

		candidate=list(skipgrams(can,2,n))
		can_sent=[Ko_tokenize.word(tmp,self.lang) for tmp in can_sent]
		can_sk_len=0

		for tmp in ref_stoken:
			ref=list(skipgrams(tmp,2,n))
			intersect=[t for t in ref if t in candidate]
			numer.append(len(intersect))
			ref_len+=len(ref)

		for tmp in can_sent:
			can_sk_len+=len(list(skipgrams(tmp,2,n)))

		prec=sum(numer)/can_sk_len
		rec=sum(numer)/ref_len

		if(prec!=0 and rec!=0):
			score = ((1 + beta**2)*prec*rec)/float(rec + beta**2*prec)
		else:
			score = 0.0
		return score


	def rouge_s(self, ref, can, n):

		can_sent= self._hyp_sent_split_remove(can)
		can_word=list(itertools.chain(*[Ko_tokenize.word(tmp,self.lang) for tmp in can_sent]))
		ref= self._ref_sent_split_remove(ref)


		r_list=[]

		for tmp in ref:
			#tmp=list(itertools.chain(*tmp))
			r_list.append(self._skip_bigrams(tmp,can_sent,can_word,n))

		return max(r_list)


	def cider(self, ref, hyp):

		ref_dict=dict()
		hyp_dict=dict()

		ref_dict[0]=ref
		hyp_dict[0]=hyp

		cider_score=Cider()
		score=cider_score.compute_score(ref_dict,hyp_dict)

		return float(score)

	def _process_espresso_output_format(self, result_list):
		temp_list = []
		for k in result_list:
			#k = k.split('_')
			k = list(k)
			if k[1] == 'SP' or k[1] == 'SY':
				continue
			temp_list.append(k)
		return temp_list

	def _generate_enum(self, ref, hyp):
		result_hyp = []
		result_ref = []
		for h in hyp:
			enum_hyp_list = list(enumerate(h))
			result_hyp.append(enum_hyp_list)
		for r in ref:
			enum_ref_list = list(enumerate(r))
			result_ref.append(enum_ref_list)
		return result_hyp, result_ref

	def _tag_pos_meteor(self, sent_list):
		result_list = list()
		for sent in sent_list:
			tagged_sent = EspressoTagger(task='pos').tag(sent)
			tagged_sent = self._process_espresso_output_format(tagged_sent)
			result_list.append(tagged_sent)
		return result_list

	def _match_enums(self,
		enum_hypothesis_list: List[Tuple[int, str]],
		enum_reference_list: List[Tuple[int, str]],
	) -> Tuple[List[Tuple[int, int]], List[Tuple[int, str]], List[Tuple[int, str]]]:
		"""
		matches exact words in hypothesis and reference and returns
		a word mapping between enum_hypothesis_list and enum_reference_list
		based on the enumerated word id.

		:param enum_hypothesis_list: enumerated hypothesis list
		:param enum_reference_list: enumerated reference list
		:return: enumerated matched tuples, enumerated unmatched hypothesis tuples,
				enumerated unmatched reference tuples
		"""
		word_match = []
		# print("test 213" , enum_hypothesis_list)
		# print("test 124" , enum_reference_list)
		for i in range(len(enum_hypothesis_list))[::-1]:
			for j in range(len(enum_reference_list))[::-1]:
				# print(f"\n \t {enum_hypothesis_list[i][1]} \t {enum_reference_list[j][1]}")
				if enum_hypothesis_list[i][1] == enum_reference_list[j][1]:

					# print("Check!!")
					word_match.append(
						(enum_hypothesis_list[i][0], enum_reference_list[j][0])
					)
					enum_hypothesis_list.pop(i)
					enum_reference_list.pop(j)
					break
		return word_match, enum_hypothesis_list, enum_reference_list


	def _count_chunks(self, matches: List[Tuple[int, int]]) -> int:
		"""
		Counts the fewest possible number of chunks such that matched unigrams
		of each chunk are adjacent to each other. This is used to calculate the
		fragmentation part of the metric.

		:param matches: list containing a mapping of matched words (output of align_words)
		:return: Number of chunks a sentence is divided into post alignment
		"""
		i = 0
		chunks = 1
		while i < len(matches) - 1:
			if (matches[i + 1][0] == matches[i][0] + 1) and (
				matches[i + 1][1] == matches[i][1] + 1
			):
				i += 1
				continue
			i += 1
			chunks += 1
		return chunks

	def _match_syn_with_sejong(self, hyp_list, ref_list):
		print("2")
		syn_match = []
		for i in range(len(hyp_list))[::-1]:
			temp_syn_list = []
			#print("test 344434: ", hyp_list[i])
			if hyp_list[i][1][1] not in self.skip_pos:
				entrys = ssem.entrys(hyp_list[i][1][0])
				print("1")
				print(entrys)
				for entry in entrys:
					print(entry)
					for sense in entry.senses():
						if sense.syn():
							temp_syn_list.append(sense.syn())
			if temp_syn_list:
				hyp_list[i][1].append(deepcopy(temp_syn_list))

			for j in range(len(ref_list))[::-1]:
				is_break = False
				if len(hyp_list[i][1]) == 3:
					for syn in hyp_list[i][1][2]:

						if syn[0] == ref_list[j][1][0]:
							syn_match.append(
								(hyp_list[i][0], ref_list[j][0])
							)
							is_break = True
							hyp_list.pop(i)
							ref_list.pop(j)
							break
				else:
					if hyp_list[i][1] == ref_list[1][1]:
						syn_match.append(
							(hyp_list[i][0], ref_list[j][0])
						)
						is_break = True
						hyp_list.pop(i)
						ref_list.pop(j)
				if is_break:
					break



			# print("test 231232 ", hyp_list[i])


		return syn_match, hyp_list, ref_list

	def meteor(self, ref, hyp):
		ref_tag = self._tag_pos_meteor(ref)
		hyp_tag = self._tag_pos_meteor(hyp)
		meteors = []
		alpha = 0.9
		beta = 3.0
		gamma = 0.5
		enum_hyp, enum_ref = self._generate_enum(ref_tag, hyp_tag)
		print("$")
		# print("test 13333 ", enum_hyp)
		for reference in enum_ref:
			hyp_len = len(enum_hyp[0])
			ref_len = len(reference)

			# 단어/어간 매칭
			word_match, enum_hyp_list, enum_ref_list = self._match_enums(deepcopy(enum_hyp[0]), reference)
			syn_match, enum_hyp_list, enum_ref_list = self._match_syn_with_sejong(enum_hyp_list, enum_ref_list)
			# print("test 123344 " ,enum_ref_list) ## [(0, ['오늘', 'NN']), (6, ['이', 'VB']), (7, ['었다', 'EE'])]

			final_match = sorted(word_match + syn_match)

			#최종 결과 계산
			final_match_count = len(final_match)


			precision = float(final_match_count) / hyp_len
			recall = float(final_match_count) / ref_len
			fmean = (precision * recall) / (alpha * precision + (1 - alpha) * recall)
			chunk_count = float(self._count_chunks(final_match))
			frag = 0.0
			if final_match_count != 0:
				frag = chunk_count / final_match_count
			else:
				frag = 0.0
			penalty = gamma * frag ** beta
			meteors.append((1 - penalty) * fmean)

			# print(word_match)

		return max(meteors)


def demo():
	y_pred = [5, 2, 4, 1, 3, 2, 5, 6, 7]
	y_true = [1, 3, 6, 7, 1, 5]

	user = [[5, 3, 2], [9, 1, 2], [3, 5, 6], [7, 2, 1]]
	h_pred = [[15, 6, 21, 3], [15, 77, 23, 14], [51, 23, 21, 2], [53, 2, 1, 5]]

	metric = DefaultMetric()
	print(metric.precision_at_k(y_true,  y_pred, 3))
	print(metric.recall_at_k(y_true,y_pred, 3))
	print(metric.hit_rate_at_k(user, h_pred, 2))

	hyp='봉준호 감독이 아카데미에서 국제영화상을 수상했다.'
	ref=['봉준호가 아카데미에서 각본상을 탔다.']
	re = metric.meteor(ref, hyp)
	print(re)

if __name__=="__main__":
	demo()



"""
string2string code
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

"""
This module implements the word embeddings class.
"""
# from tqdm import tqdm
import numpy as np
from typing import List, Union
#import torch
import os
from nltkor.make_requirement import make_requirement
try:
    import torch
    from torch import Tensor
    from torch.nn import functional as F
    import fasttext
    import fasttext.util
except ImportError:
    requirement = ['torch', 'fasttext']
    file_path = make_requirement(requirement)
    raise Exception(f"""
    Need to install Libraries, please pip install below libraries
    \t pip install torch
    \t pip install fasttext-wheel
    Or, use pip install requirement.txt
    \t  pip install -r {file_path}
    """)
# for dev purposes
import sys
# sys.path.append("/Users/dowon/nltk_ko/nltk/misc")
from nltkor.misc.string2string_default_tokenizer import Tokenizer
# from string2string_default_tokenizer import Tokenizer


class NeuralEmbeddings:
    """
    This class is an abstract class for neural word embeddings.
    """

    def __init__(self,
        tokenizer: Tokenizer = None,
    ) -> None:
        """
        Constructor.

        Arguments:
            tokenizer (Tokenizer): The tokenizer to use.
        """
        # Set the tokenizer
        if tokenizer is None:
            self.tokenizer = Tokenizer(word_delimiter=" ")



    def __call__(self,
        tokens: Union[List[str], str],
        ) -> Tensor:
        """
        This function returns the embeddings of the given tokens.

        Arguments:
            tokens (Union[List[str], str]): The tokens to embed.

        Returns:
            Tensor: The embeddings of the given tokens.
        """
        # Check the tokens
        if isinstance(tokens, str):
            tokens = self.tokenizer.tokenize(tokens)

        # Embed the tokens
        return self.embedding_layer(torch.tensor([self.vocabulary_dict[token] for token in tokens]))


    def get_embedding(self,
        tokens: Union[List[str], str]
        ) -> Tensor:
        """
        This function returns the embeddings of the given tokens.

        Arguments:
            tokens (Union[List[str], str]): The tokens to embed.

        Returns:
            Tensor: The embeddings of the given tokens.
        """
        return self.__call__(tokens)


# GloVe embeddings class
class GloVeEmbeddings(NeuralEmbeddings):
    """
    This class implements the GloVe word embeddings.
    """
    # Pre-trained GloVe embeddings
    # Source: https://github.com/stanfordnlp/GloVe#download-pre-trained-word-vectors
    MODEL_OPTIONS = {
        'glove.6B.200d': {
            'Description': 'Wikipedia 2014 + Gigaword 5 (6B tokens, 400K vocab, uncased, 300d vectors, 822 MB download)',
            'URL': 'https://huggingface.co/stanfordnlp/glove/resolve/main/glove.6B.zip',
            },
        'glove.twitter.27B': {
            'Description': 'Twitter (27B tokens, 1.2M vocab, uncased, 200d vectors, 1.42 GB download)',
            'URL': 'https://huggingface.co/stanfordnlp/glove/resolve/main/glove.twitter.27B.zip',
        },
        'glove.42B.300d': {
            'Description': 'Common Crawl (42B tokens, 1.9M vocab, uncased, 300d vectors, 1.75 GB download)',
            'URL': 'https://huggingface.co/stanfordnlp/glove/resolve/main/glove.42B.300d.zip',
        },
        'glove.840B.300d': {
            'Description': 'Common Crawl (840B tokens, 2.2M vocab, cased, 300d vectors, 2.03 GB download)',
            'URL': 'https://huggingface.co/stanfordnlp/glove/resolve/main/glove.840B.300d.zip',
        },
    }

    def __init__(self,
        model: str = 'glove.6B.200D',
        dim: int = 50,
        force_download: bool = False,
        dir = None,
        tokenizer: Tokenizer = None,
        ) -> None:
        r"""
        This function initializes the GloVe embeddings class.

        Arguments:
            model (str): The model to use. Default is 'glove.6B.200D'. (Options are: 'glove.6B.200D', 'glove.twitter.27B', 'glove.42B.300d', 'glove.840B.300d'.)
            dim (int): The dimension of the embeddings. Default is 300.
            force_download (bool): Whether to force download the model. Default is False.
            dir (str): The directory to save or load the model. Default is None.
            tokenizer (Tokenizer): The tokenizer to use. Default is None.

        Returns:
            None

        Raises:
            ValueError: If the model is not in the MODEL_OPTIONS [glove.6B.200D', 'glove.twitter.27B', 'glove.42B.300d', 'glove.840B.300d'].


        .. attention::

            If you use this class, please make sure to cite the following paper:

            .. code-block:: latex

                 @inproceedings{pennington2014glove,
                    title={Glove: Global vectors for word representation},
                    author={Pennington, Jeffrey and Socher, Richard and Manning, Christopher D},
                    booktitle={Proceedings of the 2014 conference on empirical methods in natural language processing (EMNLP)},
                    pages={1532--1543},
                    year={2014}
                }


        .. note::
            * If directory is None, the model will be saved in the torch hub directory.
            * If the model is not downloaded, it will be downloaded automatically.
        """
        # Check model
        if model not in self.MODEL_OPTIONS:
            raise ValueError(f'Invalid model: {model}.')

        # Set the attributes
        self.model = model
        self.force_download = force_download
        self.dir = dir
        self.token_size = self.model.split('.')[1]
        self.dim = dim

        # Set the path
        if self.dir is None:
            self.dir = f'{torch.hub.get_dir()}/{self.model}'

        # Remove the trailing slash
        if self.dir[-1] == '/':
            self.dir = self.dir[:-1]

        # Download the embeddings if they do not exist or if force_download is True
        if not os.path.exists(self.dir) or self.force_download:

            # Create the directory if it does not exist
            if not (os.path.exists(self.dir)):
                os.system(f'mkdir {self.dir}')

            # Download the glove .zip file
            print(f'Downloading the {self.model} zip file...')
            torch.hub.download_url_to_file(
                url=self.MODEL_OPTIONS[self.model]['URL'],
                dst=f'{self.dir}/glove.zip',
            )

            # Unzip the glove .txt files
            print(f'Unzipping the {self.model} zip file...')
            os.system(f'unzip {self.dir}/glove.zip -d {self.dir}')

            # Delete the zip file
            os.system(f'rm {self.dir}/glove.zip')

            # Process each glove .txt file and save it as a .pt file
            for file in os.listdir(self.dir):
                # Extract the words and the embeddings from the glove .txt file and save them as a .pt file

                # Example of a glove .txt file:
                # the 0.418 0.24968 -0.41242 0.1217 ...
                # ...
                # and 0.26818 0.14346 -0.27877 0.016257 ...
                # ...

                print(f'Processing {file}...')

                # Load the file
                with open(f'{self.dir}/{file}', 'r') as f:
                    lines = f.readlines()

                # Extract the dimension of the embeddings from the file name (e.g. glove.6B.200d.txt -> 200)
                file_embed_dim = file.split('.')[2][:-1]

                # Extract the words and the embeddings
                words = []
                embeddings = np.zeros((len(lines), int(file_embed_dim)))
                for i, line in enumerate(lines):
                    line = line.split(' ')
                    words.append(line[0])
                    embeddings[i] = np.array([float(x) for x in line[1:]])

                # Convert the embeddings to a tensor
                embeddings = torch.from_numpy(embeddings)

                # Save the words and the embeddings as a .pt file
                torch.save(words, f'{self.dir}/{file[:-4]}.words.pt')
                torch.save(embeddings, f'{self.dir}/{file[:-4]}.embeddings.pt')

            # Delete the glove .txt files
            os.system(f'rm -r {self.dir}/*.txt')

            # Load the weights and the vocabulary
            weights = torch.load(f'{self.dir}/glove.{self.token_size}.{self.dim}d.embeddings.pt')
            vocabulary = torch.load(f'{self.dir}/glove.{self.token_size}.{self.dim}d.words.pt')

        # If the embeddings already exist
        else:
            # Load the weights and the vocabulary
            weights = torch.load(f'{self.dir}/glove.{self.token_size}.{self.dim}d.embeddings.pt')
            vocabulary = torch.load(f'{self.dir}/glove.{self.token_size}.{self.dim}d.words.pt')

        # Create the vocabulary dictionary to be fed to the embedding layer
        self.vocabulary_dict = {word: i for i, word in enumerate(vocabulary)}

        # Create the embedding layer
        self.embedding_layer = torch.nn.Embedding.from_pretrained(
            embeddings=weights,
            freeze=True,
        )

        # Set the tokenizer
        if tokenizer is None:
            self.tokenizer = Tokenizer()
        else:
            self.tokenizer = tokenizer


    def __call__(self,
        tokens: Union[List[str], str],
        ) -> Tensor:
        """
        This function returns the embeddings of the given tokens.

        Arguments:
            tokens (Union[List[str], str]): The tokens to embed.

        Returns:
            Tensor: The embeddings of the given tokens.
        """
        return super().__call__(tokens)


    def get_embedding(self,
        tokens: Union[List[str], str]
        ) -> Tensor:
        r"""
        This function returns the embeddings of the given tokens.

        Arguments:
            tokens (Union[List[str], str]): The tokens to embed.

        Returns:
            Tensor: The embeddings of the given tokens.
        """
        return self.__call__(tokens)


# FastTextEmbeddings class
class FastTextEmbeddings(NeuralEmbeddings):
    """
    This class implements the FastText embeddings.
    """
    def __init__(self,
        model: str = 'cc.en.300.bin',
        force_download: bool = True,
        dir: str = None,
        ) -> None:
        r"""
        This function initializes the FastTextEmbeddings class.

        Arguments:
            model (str): The model to use. Some of the available models are:

                - 'cc.en.300.bin': The English model trained on Common Crawl (300 dimensions)
                - 'cc.hi.300.bin': The Hindi model trained on Common Crawl (300 dimensions)
                - 'cc.fr.300.bin': The French model trained on Common Crawl (300 dimensions)
                - 'cc.yi.300.bin': The Yiddish model trained on Common Crawl (300 dimensions)
                -  ...
                - 'wiki.en': The English model trained on Wikipedia (300 dimensions)
                - 'wiki.simple': The Simple English model trained on Wikipedia (300 dimensions)
                - 'wiki.ar': The Arabic model trained on Wikipedia (300 dimensions)
                - 'wiki.bg': The Bulgarian model trained on Wikipedia (300 dimensions)
                - 'wiki.ca': The Catalan model trained on Wikipedia (300 dimensions)
                - 'wiki.zh': The Chinese model trained on Wikipedia (300 dimensions)
                - 'wiki.sw': The Swahili model trained on Wikipedia (300 dimensions)
                - 'wiki.fr': The French model trained on Wikipedia (300 dimensions)
                - 'wiki.de': The German model trained on Wikipedia (300 dimensions)
                - 'wiki.es': The Spanish model trained on Wikipedia (300 dimensions)
                - 'wiki.it': The Italian model trained on Wikipedia (300 dimensions)
                - 'wiki.pt': The Portuguese model trained on Wikipedia (300 dimensions)
                - 'wiki.ru': The Russian model trained on Wikipedia (300 dimensions)
                - 'wiki.tr': The Turkish model trained on Wikipedia (300 dimensions)
                - 'wiki.uk': The Ukrainian model trained on Wikipedia (300 dimensions)
                - 'wiki.vi': The Vietnamese model trained on Wikipedia (300 dimensions)
                - 'wiki.id': The Indonesian model trained on Wikipedia (300 dimensions)
                - 'wiki.ja': The Japanese model trained on Wikipedia (300 dimensions)
                - ...

            force_download (bool): Whether to force the download of the model. Default: False.
            dir (str): The directory to save and load the model.

        Returns:
            None

        Raises:
            ValueError: If the given model is not available.

        .. attention::

            If you make use of this code, please cite the following papers (depending on the model you use):

            .. code-block:: latex

                @inproceedings{mikolov2018advances,
                    title={Advances in Pre-Training Distributed Word Representations},
                    author={Mikolov, Tomas and Grave, Edouard and Bojanowski, Piotr and Puhrsch, Christian and Joulin, Armand},
                    booktitle={Proceedings of the International Conference on Language Resources and Evaluation (LREC 2018)},
                    year={2018}
                }

            .. code-block:: latex

                @article{bojanowski2017enriching,
                    title={Enriching Word Vectors with Subword Information},
                    author={Bojanowski, Piotr and Grave, Edouard and Joulin, Armand and Mikolov, Tomas},
                    journal={Transactions of the Association for Computational Linguistics},
                    volume={5},
                    year={2017},
                    issn={2307-387X},
                    pages={135--146}
                }

            .. code-block:: latex

                @article{joulin2016fasttext,
                    title={FastText.zip: Compressing text classification models},
                    author={Joulin, Armand and Grave, Edouard and Bojanowski, Piotr and Douze, Matthijs and J{\'e}gou, H{\'e}rve and Mikolov, Tomas},
                    journal={arXiv preprint arXiv:1612.03651},
                    year={2016}
                }

        .. note::

            * The models are downloaded from https://fasttext.cc/docs/en/english-vectors.html.
            * The models are saved in the torch hub directory, if no directory is specified.
            *
        """

        # Set the attributes
        self.model = model
        self.dir = dir
        self.force_download = force_download

        # Set the path
        if self.dir is None:
            # For convenience, we save the model in the torch hub directory
            self.dir = f'{torch.hub.get_dir()}/{self.model}'

        # Remove the trailing slash
        if self.dir[-1] == '/':
            self.dir = self.dir[:-1]

        # Download the embeddings if they do not exist or if force_download is True
        if not os.path.exists(self.dir) or self.force_download:
            # Create the directory if it does not exist
            if not os.path.exists(self.dir):
                os.system(f'mkdir {self.dir}')

            # Download using wget
            if 'wiki' in model:
                # https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/wiki.en.zip
                os.system(f'wget https://dl.fbaipublicfiles.com/fasttext/vectors-wiki/{model}.zip -P {self.dir}')
                os.system(f'unzip {self.dirl}.zip -d {self.dir}')
                os.system(f'rm {self.dir}.zip')
            else:
                # https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/cc.en.300.bin.gz
                os.system(f'wget https://dl.fbaipublicfiles.com/fasttext/vectors-crawl/{model}.gz -P {self.dir}')
                os.system(f'gunzip {self.dir}.gz -d {self.dir}')
                os.system(f'rm {self.dir}.gz')

            # Load the model
            ft = fasttext.load_model(f'{self.dir}/{model}')

            # Get the vocabulary
            words = ft.get_words()

            # Convert the embeddings to a tensor
            embeddings =torch.tensor(ft.get_input_matrix())

            # Save the words and the embeddings as a .pt file
            torch.save(words, f'{self.dir}/{model}.words.pt')
            torch.save(embeddings, f'{self.dir}/{model}.embeddings.pt')

            # Delete the model
            del ft

        else:
            try:
                # Load the words and the embeddings
                words = torch.load(f'{self.dir}/{model}.words.pt')
                embeddings = torch.load(f'{self.dir}/{model}.embeddings.pt')
            except:
                raise Exception(f'Please install the {model} model first by setting force_download to True.')

        # Create the vocabulary dictionary to be fed to the embedding layer
        self.vocabulary_dict = {word: i for i, word in enumerate(words)}

        # Create the embedding layer
        self.embedding_layer = torch.nn.Embedding.from_pretrained(
            embeddings=embeddings,
            freeze=True,
        )

    def __call__(self,
        tokens: Union[List[str], str],
        ) -> Tensor:
        """
        This function returns the embeddings of the given tokens.

        Arguments:
            tokens (Union[List[str], str]): The tokens to embed.

        Returns:
            Tensor: The embeddings of the given tokens.
        """
        return super().__call__(tokens)


    def get_embedding(self,
        tokens: Union[List[str], str]
        ) -> Tensor:
        """
        This function returns the embeddings of the given tokens.

        Arguments:
            tokens (Union[List[str], str]): The tokens to embed.

        Returns:
            Tensor: The embeddings of the given tokens.
        """
        return self.__call__(tokens)



# -*- coding: utf-8 -*-

"""
Code for argument parsing and a few verifications. 
These arguments are used by the training script.
"""

import argparse


def fill_defaults(args, defaults_per_task):
		"""
		This function fills arguments not explicitly set (left as None)
		with default values according to the chosen task.
		
		We can't rely on argparse to it because using subparsers with
		set_defaults and a parent parser overwrites the defaults. 
		"""
		task = args.task
		defaults = defaults_per_task[task]
		for arg in args.__dict__:
				if getattr(args, arg) is None and arg in defaults:
						setattr(args, arg, defaults[arg])
		

def get_args():
		parser = argparse.ArgumentParser(description="Train the Espresso "\
																		 "for a given NLP task.")
		subparsers = parser.add_subparsers(title='Tasks',
																			 dest='task',
																			 description='Task to train the Espresso for. '\
																			 'Type %(prog)s [TASK] -h to get task-specific help.')
		
		defaults = {}
		
		# base parser with arguments not related to any model
		base_parser = argparse.ArgumentParser(add_help=False)
		base_parser.add_argument('-f', '--num_features', type=int,
														 help='Number of features per word '\
														 '(used to generate random vectors)',
														 default=50, dest='num_features')
		base_parser.add_argument('--load_network', action='store_true',
														 help='Load previously saved network')
		base_parser.add_argument('--load_features', action='store_true',
														 help="Load previously saved word type features "\
														 "(overrides -f and must also load a vocabulary file)", 
														 dest='load_types')
		base_parser.add_argument('-v', '--verbose', help='Verbose mode',
														 action="store_true")
		base_parser.add_argument('--gold', help='File with annotated data for training.', 
														 type=str, required=True)
		base_parser.add_argument('--data', help='Directory to save new models and load '\
														 'partially trained ones (default: current directory)', type=str, 
														 default='./data/')
		base_parser.add_argument('--dev', help='Development (validation) data. If not given, '\
														 'training data will be used to evaluate performance.',
														 default=None)
		
		# parser with network arguments shared among most tasks
		# each task-specific parser may define defaults
		network_parser = argparse.ArgumentParser(add_help=False, parents=[base_parser])
		
		network_parser.add_argument('-w', '--window', type=int,
																help='Size of the word window',
																dest='window')	# 반드시 홀수로 할 것 	
		network_parser.add_argument('-e', '--epochs', type=int, dest='iterations',
																help='Number of training epochs')
		network_parser.add_argument('-l', '--learning_rate', type=float,
																help='Learning rate for network connections',
																dest='learning_rate')
		network_parser.add_argument('--lf', type=float,
																help='Learning rate for features',
																dest='learning_rate_features')
		network_parser.add_argument('--lt', type=float,
																help='Learning rate for tag transitions',
																dest='learning_rate_transitions')
		network_parser.add_argument('--decay', type=float, const=1, nargs='?', default=None,
																help='Use learning rate decay. Optionally, '\
																'supply decay factor (default 1)')
		network_parser.add_argument('-a', '--accuracy', type=float,
																help='Maximum desired accuracy per token.',
																default=0, dest='accuracy')
		network_parser.add_argument('-n', '--hidden', type=int,
																help='Number of hidden neurons',
																dest='hidden')
		network_parser.add_argument('--caps', const=5, nargs='?', type=int, default=None,
																help='Include capitalization features. '\
																'Optionally, supply the number of features (default 5)')
		
				
		# parser with arguments shared among convolutional-based tasks
		conv_parser = argparse.ArgumentParser(add_help=False)
		conv_parser.add_argument('-c', '--convolution', type=int, 
														 help='Number of convolution neurons',
														 dest='convolution')
		conv_parser.add_argument('--pos', const=5, nargs='?', type=int, default=None,
														 help='Include part-of-speech features. '\
														 'Optionally, supply the number of features (default 5)')
		conv_parser.add_argument('--max_dist', type=int, default=10,
														 help='Maximum distance to have its own feature vector')
		conv_parser.add_argument('--target_features', type=int, default=5,
														 help='Number of features for distance to target word')
		conv_parser.add_argument('--pred_features', type=int, default=5,
														 help='Number of features for distance to predicate')
				
		
		# POS argument parser
		parser_pos = subparsers.add_parser('pos', help='POS tagging', 
																			 parents=[network_parser])
		parser_pos.add_argument('--suffix', const=2, nargs='?', type=int, default=None,
														help='Include suffix features. Optionally, '\
														'supply the number of features (default 2)')
		parser_pos.add_argument('--suffix_size', type=int, default=5,
														help='Use suffixes up to this size (in characters, default 5). '\
														'Only used if --suffix is supplied')
		parser_pos.add_argument('--prefix', const=2, nargs='?', type=int, default=None,
														help='Include prefix features. Optionally, '\
														'supply the number of features (default 2)')
		parser_pos.add_argument('--prefix_size', type=int, default=5,
														help='Use prefixes up to this size (in characters, default 5). '\
														'Only used if --suffix is supplied')
		defaults['pos'] = dict(window=9, hidden=300, iterations=20, 
													 learning_rate=0.001, learning_rate_features=0.001,
													 learning_rate_transitions=0.001 )

		# NER argument parser
		parser_ner = subparsers.add_parser('ner', help='Named Entity Tagging', 
																			 parents=[network_parser])
		parser_ner.add_argument('--suffix', const=2, nargs='?', type=int, default=None,
														help='Include suffix features. Optionally, '\
														'supply the number of features (default 2)')
		parser_ner.add_argument('--suffix_size', type=int, default=5,
														help='Use suffixes up to this size (in characters, default 5). '\
														'Only used if --suffix is supplied')
		parser_ner.add_argument('--prefix', const=2, nargs='?', type=int, default=None,
														help='Include prefix features. Optionally, '\
														'supply the number of features (default 2)')
		parser_ner.add_argument('--prefix_size', type=int, default=5,
														help='Use prefixes up to this size (in characters, default 5). '\
														'Only used if --suffix is supplied')
		defaults['ner'] = dict(window=5, hidden=300, iterations=20, 
													 learning_rate=0.001, learning_rate_features=0.001,
													 learning_rate_transitions=0.001)

		# WSD argument parser
		parser_wsd = subparsers.add_parser('wsd', help='Word Sense Disambiguation', 
																			 parents=[network_parser])
		parser_wsd.add_argument('--suffix', const=2, nargs='?', type=int, default=None,
														help='Include suffix features. Optionally, '\
														'supply the number of features (default 2)')
		parser_wsd.add_argument('--suffix_size', type=int, default=5,
														help='Use suffixes up to this size (in characters, default 5). '\
														'Only used if --suffix is supplied')
		parser_wsd.add_argument('--prefix', const=2, nargs='?', type=int, default=None,
														help='Include prefix features. Optionally, '\
														'supply the number of features (default 2)')
		parser_wsd.add_argument('--prefix_size', type=int, default=5,
														help='Use prefixes up to this size (in characters, default 5). '\
														'Only used if --suffix is supplied')
		defaults['wsd'] = dict(window=5, hidden=300, iterations=20, 
													 learning_rate=0.001, learning_rate_features=0.001,
													 learning_rate_transitions=0.001)

	# Chunk argument parser
		#parser_ner = subparsers.add_parser('ner', help='Named Entity Tagging', 
		#																	 parents=[network_parser])
		#parser_ner.add_argument('--suffix', const=2, nargs='?', type=int, default=None,
		#												help='Include suffix features. Optionally, '\
		#												'supply the number of features (default 2)')
		#parser_ner.add_argument('--suffix_size', type=int, default=5,
		#												help='Use suffixes up to this size (in characters, default 5). '\
		#												'Only used if --suffix is supplied')
		#parser_ner.add_argument('--prefix', const=2, nargs='?', type=int, default=None,
		#												help='Include prefix features. Optionally, '\
		#												'supply the number of features (default 2)')
		#parser_ner.add_argument('--prefix_size', type=int, default=5,
		#												help='Use prefixes up to this size (in characters, default 5). '\
		#												'Only used if --suffix is supplied')
		#defaults['ner'] = dict(window=7, hidden=300, iterations=20, 
		#											 learning_rate=0.001, learning_rate_features=0.001,
		#											 learning_rate_transitions=0.001)

		
		# dependency
		parser_dep = subparsers.add_parser('dependency', help='Dependency parsing')
		dep_subparsers = parser_dep.add_subparsers(title='Dependency parsing training steps',
																							 dest='subtask',
																							 description='Which step of the dependency training '\
																							 '(detecting edges or labeling them)')
		
		dep_subparsers.add_parser('labeled', help='Labeling dependency edges',
															parents=[network_parser, conv_parser])
		dep_subparsers.add_parser('unlabeled', help='Dependency edge detection',
															parents=[network_parser, conv_parser])
		
		defaults['dependency_filter'] = dict()
		defaults['labeled_dependency'] = dict(window=9, hidden=200, iterations=20, 
													 learning_rate=0.001, learning_rate_features=0.001,
													 convolution=300, pos=50,
													 learning_rate_transitions=0.001)
		defaults['unlabeled_dependency'] = dict(window=9, hidden=200, iterations=20, 
													 learning_rate=0.001, learning_rate_features=0.001,
													 convolution=300, pos=50,
													 learning_rate_transitions=0.001)
		

		# SRL argument parser
		# There is another level of subparsers for predicate detection / 
		# argument boundary identification / argument classification / 
		# (id + class) in one step
		parser_srl = subparsers.add_parser('srl', help='Semantic Role Labeling',
																			parents=[network_parser, conv_parser],
																			 formatter_class=argparse.RawDescriptionHelpFormatter)
		#parser_srl.set_defaults(identify=False, predicates=False, classify=False)

		desc = '''SRL has 3 steps: predicate	detection, argument identification and 
argument classification. Each one depends on the one before.

You need one model trained for each subtask (or one for predicate
detection and another with the other 2 steps) in order to perform
full SRL.

Type %(prog)s [SUBTASK] -h to get subtask-specific help.'''
		
		'''
		srl_subparsers = parser_srl.add_subparsers(title='SRL subtasks',
																							 dest='subtask',
																							 description=desc)
		srl_subparsers.add_parser('pred', help='Predicate identification',
															parents=[network_parser])
		defaults['srl_predicates'] = dict(window=5, hidden=50, iterations=1, 
																			learning_rate=0.01, learning_rate_features=0.01,
																			learning_rate_transitions=0.01,
																			predicates=True)
		
		srl_subparsers.add_parser('id', help='Argument identification',
															parents=[network_parser, conv_parser])
		defaults['srl_boundary'] = dict(window=3, hidden=150, convolution=150, 
																		identify=True, iterations=15,
																		learning_rate=0.001, learning_rate_features=0.001,
																		learning_rate_transitions=0.001)
		
		srl_subparsers.add_parser('class', help='Argument classification',
															parents=[network_parser, conv_parser])
		defaults['srl_classify'] = dict(window=3, hidden=0, convolution=100, 
																		classify=True, iterations=3,
																		learning_rate=0.01, learning_rate_features=0.01,
																		learning_rate_transitions=0.01)
		srl_subparsers.add_parser('1step', parents=[network_parser, conv_parser],
															help='Argument identification and '\
															'classification together')
		'''
		defaults['srl'] = dict(window=3, hidden=150, convolution=200, iterations=8,
														pos = 5,
													 learning_rate=0.001, learning_rate_features=0.001,
													 learning_rate_transitions=0.001)
		
		
		args = parser.parse_args()
		
		#if args.task == 'srl':
			#	if args.subtask == 'class':
			#			args.task = 'srl_classify'
			#			args.classify = True
			#	elif args.subtask == 'id':
			#			args.task = 'srl_boundary'
			#			args.identify = True
			#	elif args.subtask == 'pred':
			#			args.task = 'srl_predicates'
			#			args.predicates = True
		if args.task == 'dependency':
			if args.subtask == 'labeled':
				args.task = 'labeled_dependency'
				args.labeled = True
			elif args.subtask == 'unlabeled':
				args.task = 'unlabeled_dependency'
				args.labeled = False
				
		fill_defaults(args, defaults)
		return args



"""
string2string code
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

"""
This file contains the default tokenizer.
"""

from typing import List

# Tokenizer class
class Tokenizer:
    """
    This class contains the tokenizer.
    """

    def __init__(self,
        word_delimiter: str = " ",
    ):
        """
        Initializes the Tokenizer class.

        Arguments:
            word_delimiter (str): The word delimiter. Default is " ".
        """
        # Set the word delimiter
        self.word_delimiter = word_delimiter

    # Tokenize
    def tokenize(self,
        text: str,
    ) -> List[str]:
        """
        Returns the tokens from a string.

        Arguments:
            text (str): The text to tokenize.

        Returns:
            List[str]: The tokens.
        """
        return text.split(self.word_delimiter)
    
    # Detokenize
    def detokenize(self,
        tokens: List[str],
    ) -> str:
        """
        Returns the string from a list of tokens.

        Arguments:
            tokens (List[str]): The tokens.

        Returns:
            str: The string.
        """
        return self.word_delimiter.join(tokens)


from collections import defaultdict
from nltk.translate.bleu_score import *
from nltk.metrics import scores
from nltk.metrics import confusionmatrix
from nltkor.tokenize import word_tokenize,sent_tokenize,syllable_tokenize
from nltk.util import ngrams, skipgrams
from nltkor.cider.cider import Cider
import sys
import itertools
from nltkor.tag import pos_tag, pos_tag_with_verb_form, EspressoTagger
from nltkor.sejong import ssem
from typing import Callable, Iterable, List, Tuple
from copy import deepcopy

# Natural Language Toolkit: Machine Translation
#
# Copyright (C) 2001-2023 NLTK Project
# Author: Uday Krishna <udaykrishna5@gmail.com>
# Contributor: Tom Aarsen
# URL: <https://www.nltk.org/>
# For license information, see LICENSE.TXT

class StringMetric:
	def __init__(self, lang="korean"):
		self.lang = lang
		if lang not in ["korean", "english"]:
			raise Exception("Only \"korean\" or \"english\" in lang parameter")
		self.tokenize=lambda ref: word_tokenize(ref,self.lang)
		self.skip_pos = ['EE']


	def _W_CER(self, r, h):

		costs = [[0 for inner in range(len(h)+1)] for outer in range(len(r)+1)]

		DEL_PENALTY=1 # Tact
		INS_PENALTY=1 # Tact
		SUB_PENALTY=1 # Tact

		for i in range(1, len(r)+1):
			costs[i][0] = DEL_PENALTY*i

		for j in range(1, len(h) + 1):
			costs[0][j] = INS_PENALTY*j

		# computation
		for i in range(1, len(r)+1):
			for j in range(1, len(h)+1):
				if r[i-1] == h[j-1]:
					costs[i][j] = costs[i-1][j-1]
				else:
					substitutionCost = costs[i-1][j-1] + SUB_PENALTY # penalty is always 1
					insertionCost	= costs[i][j-1] + INS_PENALTY   # penalty is always 1
					deletionCost	 = costs[i-1][j] + DEL_PENALTY   # penalty is always 1

					costs[i][j] = min(substitutionCost, insertionCost, deletionCost)

		mo = len(r)
		i = len(r)
		j = len(h)

		result=(costs[i][j])/mo

		if result>1.0:
			return 1.0
		else:
			return result


	def wer(self, reference, candidate):
		r = word_tokenize(reference, self.lang)
		h = word_tokenize(candidate, self.lang)

		return self._W_CER(r,h)


	def cer(self, reference,candidate):
		r = syllable_tokenize(reference, self.lang)
		h = syllable_tokenize(candidate, self.lang)

		return self._W_CER(r,h)


	def bleu(self, reference, candidate,weights=(0.25,0.25,0.25,0.25), smoothing_function=None):

		if type(candidate)!=list or type(reference)!=list:
			print("parameter expect list type")
			return

		reference=list(map(self.tokenize,reference))
		candidate=word_tokenize(candidate,self.lang)

		return sentence_bleu(reference,candidate,weights,smoothing_function=smoothing_function)


	def bleu_n(self, reference,candiate,n=1, smoothing_function=None):

		if n==1:
			return self.bleu(reference,candiate,(1,0,0,0), smoothing_function=smoothing_function)
		elif n==2:
			return self.bleu(reference,candiate,(0,1,0,0), smoothing_function=smoothing_function)
		elif n==3:
			return self.bleu(reference,candiate,(0,0,1,0), smoothing_function=smoothing_function)
		elif n==4:
			return self.bleu(reference,candiate,(0,0,0,1), smoothing_function=smoothing_function)




	def _hyp_sent_split_remove(self, can):

		can_sent=[[tmp.rstrip('.'or'?'or'!'or','or'\n')] for tmp in sent_tokenize(can, self.lang)]
		return can_sent

	def _ref_sent_split_remove(self, ref):

		ref_sent=[sent_tokenize(tmp,self.lang) for tmp in ref]
		ref_sent_c=[]
		for tmp in ref_sent:
			ref_sent_in=[]
			for tmp2 in tmp:
				ref_sent_in.append(word_tokenize(tmp2.rstrip('.'or'?'or'!'or','or'\n'),self.lang))
			ref_sent_c.append(ref_sent_in)

		return ref_sent_c

	def _token(self, ref_stoken, can, n):

		numer=[]
		ref_len=0

		can=list(ngrams(can,n))

		for tmp in ref_stoken:

			if n==1:
				ref=list(ngrams(tmp,1))
			elif n==2:
				ref=list(ngrams(tmp,2))
			elif n==3:
				ref=list(ngrams(tmp,3))
			else:return 0

			intersect = [t for t in ref if t in can ]
			numer.append(len(intersect))
			ref_len+=len(ref)

		try:
			rec= sum(numer)/ref_len
		except:
			rec=0

		return rec



	def rouge_n(self, ref, can, n=1):

		beta=1
		rec,prec=0,0

		can_sent=self._hyp_sent_split_remove(can)
		can_word=list(itertools.chain(*[word_tokenize(tmp,self.lang) for tmp in can_sent]))
		ref=self._ref_sent_split_remove(ref)

		r_list=[]

		for tmp in ref:
			if n==1:
				r_list.append(self._token(tmp, can_word, 1))
			elif n==2:
				r_list.append(self._token(tmp, can_word, 2))
			elif n==3:
				r_list.append(self._token(tmp, can_word, 3))

		return max(r_list)



	def rouge_l(self, ref, can):

		beta=1
		#check=0

		can= self._hyp_sent_split_remove(can)
		can=[word_tokenize(tmp,self.lang) for tmp in can]
		refs=self._ref_sent_split_remove(ref)

		can_word=list(itertools.chain(*can))

		result_list=[]

		for ref in refs:
			lcs_list=[]
			for ri in ref:
				ri_C=[]
				for ci in can:
					temp=self._lcs(ci,ri)
					ri_C.append(temp)

				ri_C=list(itertools.chain(*ri_C))
				ri_C=set(ri_C)
				lcs_list.append(len(ri_C))

			ref_word=list(itertools.chain(*ref))

			R_lcs=sum(lcs_list)/len(ref_word)
			P_lcs=sum(lcs_list)/len(can_word)

			try:
				F_lcs= (2*R_lcs*P_lcs)/(R_lcs+P_lcs)
			except:
				F_lcs=0
			result_list.append(F_lcs)

		return max(result_list)



	def _lcs(self, can, ref):


		s1=can
		s2=ref
		check=0

		if len(s1)<=len(s2):
			temp=s1
			s1=s2
			s2=temp
			check=1

		m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]

		for x in range(1, 1 + len(s1)):
			for y in range(1, 1 + len(s2)):
				if s1[x - 1] == s2[y - 1]:
					m[x][y] = m[x - 1][y - 1] +1
				else:
					m[x][y]=max(m[x][y-1],m[x-1][y])
		f_x=len(s2)+1
		lcs=m[len(s1)][len(s2)]
		temp=[]


		i=len(s1)
		j=len(s2)

		while m[i][j]!=0:
			if(m[i][j]==m[i][j-1]):
				j-=1
			elif (m[i][j]==m[i-1][j]):
				i-=1
			else:
				if check==0:
					temp.append(s1[i-1])
				if check==1:
					temp.append(s2[j-1])
				i-=1
				j-=1

		return temp
		'''
		for y in reversed(range(1,1+len(s1))):
			for x in reversed(range(1,1+len(s2))):
				if (m[y][x]-m[y-1][x-1]==1) and (m[y][x]-m[y-1][x]==1) and (m[y][x]-m[y][x-1]==1):
					if (y==len(s1)+1) and (x==len(s2)):
						temp.append(x)
					else:
						temp.append(x-1)

		print('the police 만  나와줘야',temp)
		if check==0:
			word=s1
		elif check==1:
			word=s2

		ret_list=[]

		for tmp in range(len(temp)):
			ret_list.append(word[temp[tmp]])

		return ret_list
		'''


	def _skip_bigrams(self, ref_stoken, can_sent, can, n=1):

		beta=1
		numer=[]
		ref_len=0

		candidate=list(skipgrams(can,2,n))
		can_sent=[word_tokenize(tmp,self.lang) for tmp in can_sent]
		can_sk_len=0

		for tmp in ref_stoken:
			ref=list(skipgrams(tmp,2,n))
			intersect=[t for t in ref if t in candidate]
			numer.append(len(intersect))
			ref_len+=len(ref)

		for tmp in can_sent:
			can_sk_len+=len(list(skipgrams(tmp,2,n)))

		prec=sum(numer)/can_sk_len
		rec=sum(numer)/ref_len

		if(prec!=0 and rec!=0):
			score = ((1 + beta**2)*prec*rec)/float(rec + beta**2*prec)
		else:
			score = 0.0
		return score


	def rouge_s(self, ref, can, n):

		can_sent= self._hyp_sent_split_remove(can)
		can_word=list(itertools.chain(*[word_tokenize(tmp,self.lang) for tmp in can_sent]))
		ref= self._ref_sent_split_remove(ref)


		r_list=[]

		for tmp in ref:
			#tmp=list(itertools.chain(*tmp))
			r_list.append(self._skip_bigrams(tmp,can_sent,can_word,n))

		return max(r_list)


	def cider(self, ref, hyp):

		ref_dict=dict()
		hyp_dict=dict()

		ref_dict[0]=ref
		hyp_dict[0]=hyp

		cider_score=Cider()
		score=cider_score.compute_score(ref_dict,hyp_dict)

		return float(score)

	def _process_espresso_output_format(self, result_list):
		temp_list = []
		for k in result_list:
			k = k.split('_')
			if k[1] == 'SP' or k[1] == 'SY':
				continue
			temp_list.append(k)
		return temp_list

	def _generate_enum(self, ref, hyp):
		result_hyp = []
		result_ref = []
		for h in hyp:
			enum_hyp_list = list(enumerate(h))
			result_hyp.append(enum_hyp_list)
		for r in ref:
			enum_ref_list = list(enumerate(r))
			result_ref.append(enum_ref_list)
		return result_hyp, result_ref

	def _tag_pos_meteor(self, sent_list):
		result_list = list()
		for sent in sent_list:
			tagged_sent = EspressoTagger(task='pos').tag(sent)
			tagged_sent = self._process_espresso_output_format(tagged_sent)
			result_list.append(tagged_sent)
		return result_list

	def _match_enums(self,
		enum_hypothesis_list: List[Tuple[int, str]],
		enum_reference_list: List[Tuple[int, str]],
	) -> Tuple[List[Tuple[int, int]], List[Tuple[int, str]], List[Tuple[int, str]]]:
		"""
		matches exact words in hypothesis and reference and returns
		a word mapping between enum_hypothesis_list and enum_reference_list
		based on the enumerated word id.

		:param enum_hypothesis_list: enumerated hypothesis list
		:param enum_reference_list: enumerated reference list
		:return: enumerated matched tuples, enumerated unmatched hypothesis tuples,
				enumerated unmatched reference tuples
		"""
		word_match = []
		# print("test 213" , enum_hypothesis_list)
		# print("test 124" , enum_reference_list)
		for i in range(len(enum_hypothesis_list))[::-1]:
			for j in range(len(enum_reference_list))[::-1]:
				# print(f"\n \t {enum_hypothesis_list[i][1]} \t {enum_reference_list[j][1]}")
				if enum_hypothesis_list[i][1] == enum_reference_list[j][1]:

					# print("Check!!")
					word_match.append(
						(enum_hypothesis_list[i][0], enum_reference_list[j][0])
					)
					enum_hypothesis_list.pop(i)
					enum_reference_list.pop(j)
					break
		return word_match, enum_hypothesis_list, enum_reference_list


	def _count_chunks(self, matches: List[Tuple[int, int]]) -> int:
		"""
		Counts the fewest possible number of chunks such that matched unigrams
		of each chunk are adjacent to each other. This is used to calculate the
		fragmentation part of the metric.

		:param matches: list containing a mapping of matched words (output of align_words)
		:return: Number of chunks a sentence is divided into post alignment
		"""
		i = 0
		chunks = 1
		while i < len(matches) - 1:
			if (matches[i + 1][0] == matches[i][0] + 1) and (
				matches[i + 1][1] == matches[i][1] + 1
			):
				i += 1
				continue
			i += 1
			chunks += 1
		return chunks

	def _match_syn_with_sejong(self, hyp_list, ref_list):
		syn_match = []
		for i in range(len(hyp_list))[::-1]:
			temp_syn_list = []
			#print("test 344434: ", hyp_list[i])
			if hyp_list[i][1][1] not in self.skip_pos:
				entrys = ssem.entrys(hyp_list[i][1][0])
				for entry in entrys:
					for sense in entry.senses():
						if sense.syn():
							temp_syn_list.append(sense.syn())
			if temp_syn_list:
				hyp_list[i][1].append(deepcopy(temp_syn_list))

			for j in range(len(ref_list))[::-1]:
				is_break = False
				if len(hyp_list[i][1]) == 3:
					for syn in hyp_list[i][1][2]:

						if syn[0] == ref_list[j][1][0]:
							syn_match.append(
								(hyp_list[i][0], ref_list[j][0])
							)
							is_break = True
							hyp_list.pop(i)
							ref_list.pop(j)
							break
				else:
					if hyp_list[i][1] == ref_list[1][1]:
						syn_match.append(
							(hyp_list[i][0], ref_list[j][0])
						)
						is_break = True
						hyp_list.pop(i)
						ref_list.pop(j)
				if is_break:
					break



			# print("test 231232 ", hyp_list[i])


		return syn_match, hyp_list, ref_list

	def meteor(self, ref, hyp):
		ref_tag = self._tag_pos_meteor(ref)
		hyp_tag = self._tag_pos_meteor(hyp)
		meteors = []
		alpha = 0.9
		beta = 3.0
		gamma = 0.5
		enum_hyp, enum_ref = self._generate_enum(ref_tag, hyp_tag)
		# print("test 13333 ", enum_hyp)
		for reference in enum_ref:
			hyp_len = len(enum_hyp[0])
			ref_len = len(reference)

			# 단어/어간 매칭
			word_match, enum_hyp_list, enum_ref_list = self._match_enums(deepcopy(enum_hyp[0]), reference)
			syn_match, enum_hyp_list, enum_ref_list = self._match_syn_with_sejong(enum_hyp_list, enum_ref_list)
			# print("test 123344 " ,enum_ref_list) ## [(0, ['오늘', 'NN']), (6, ['이', 'VB']), (7, ['었다', 'EE'])]

			final_match = sorted(word_match + syn_match)

			#최종 결과 계산
			final_match_count = len(final_match)


			precision = float(final_match_count) / hyp_len
			recall = float(final_match_count) / ref_len
			fmean = (precision * recall) / (alpha * precision + (1 - alpha) * recall)
			chunk_count = float(self._count_chunks(final_match))
			frag = 0.0
			if final_match_count != 0:
				frag = chunk_count / final_match_count
			else:
				frag = 0.0
			penalty = gamma * frag ** beta
			meteors.append((1 - penalty) * fmean)

			# print(word_match)

		return max(meteors)


if __name__=="__main__":
	hyp='봉준호 감독이 아카데미에서 국제영화상을 수상했다.'
	ref=['봉준호가 아카데미에서 각본상을 탔다.']
	metric = StringMetric()
	re = metric.meteor(ref, hyp)
	print(re)



# Author: Krishna Pillutla
# License: GPLv3

import math
import numpy as np
import time
from types import SimpleNamespace
from nltkor.make_requirement import make_requirement

import faiss
from sklearn.preprocessing import normalize
from sklearn.decomposition import PCA
from sklearn.metrics import auc as compute_area_under_curve
import torch
from transformers import AutoModel, AutoTokenizer

try:
    import torch
    FOUND_TORCH = True
except (ImportError, ModuleNotFoundError):
    FOUND_TORCH = False

try:
    import transformers
    FOUND_TRANSFORMERS = True
except (ImportError, ModuleNotFoundError):
    FOUND_TRANSFORMERS = False

if FOUND_TORCH and FOUND_TRANSFORMERS:
    # only needed for tokenizing
    from .mauve_utils import get_tokenizer, get_model, featurize_tokens_from_model, get_device_from_arg


MODEL, TOKENIZER, MODEL_NAME = None, None, None

class Mauve:
    def __init__(self, model_name_or_path='skt/kobert-base-v1'):
        self.featurize_model_name=model_name_or_path

    def compute(self,
            p_features=None, q_features=None,
            p_tokens=None, q_tokens=None,
            p_text=None, q_text=None,
            num_buckets='auto', pca_max_data=-1, kmeans_explained_var=0.9,
            kmeans_num_redo=5, kmeans_max_iter=500,
            device_id=-1, max_text_length=1024,
            divergence_curve_discretization_size=25, mauve_scaling_factor=5,
            verbose=False, seed=25, batch_size=1, use_float64=False,
    ):

        """
        Compute the MAUVE score between two text generations P and Q.

        P is either specified as ``p_features``, ``p_tokens``, or ``p_text``. Same with Q.

        :param ``p_features``: ``numpy.ndarray`` of shape (n, d), where n is the number of generations.
        :param ``q_features``: ``numpy.ndarray`` of shape (n, d), where n is the number of generations.
        :param ``p_tokens``: list of length n, each entry is torch.LongTensor of shape (1, length).
        :param ``q_tokens``: list of length n, each entry is torch.LongTensor of shape (1, length).
        :param ``p_text``: list of length n, each entry is a string.
        :param ``q_text``: list of length n, each entry is a string.
        :param ``num_buckets``: the size of the histogram to quantize P and Q. Options: ``'auto'`` (default, which is n/10) or an integer.
        :param ``pca_max_data``: the number data points to use for PCA. If `-1`, use all the data. Default -1.
        :param ``kmeans_explained_var``: amount of variance of the data to keep in dimensionality reduction by PCA. Default 0.9.
        :param ``kmeans_num_redo``: number of times to redo k-means clustering (the best objective is kept). Default 5.
            Try reducing this to 1 in order to reduce running time.
        :param ``kmeans_max_iter``: maximum number of k-means iterations. Default 500.
            Try reducing this to 100 in order to reduce running time.
        :param ``featurize_model_name``: name of the model from which features are obtained. Default 'gpt2-large'.
            We support all models which can be loaded from ``transformers.AutoModel.from_pretrained(featurize_model_name)``.
        :param ``device_id``: Device for featurization. Supply gpu_id (e.g. 0 or 3) to use GPU or -1 to use CPU.
        :param ``max_text_length``: maximum number of tokens to consider. Default 1024.
        :param ``divergence_curve_discretization_size``: Number of points to consider on the divergence curve. Default 25.
            Larger values do not offer much of a difference.
        :param ``mauve_scaling_factor``: The constant``c`` from the paper. Default 5.
            See `Best Practices <index.html#best-practices-for-mauve>`_ for details.
        :param ``verbose``: If True, print running time updates.
        :param ``seed``: random seed to initialize k-means cluster assignments.
        :param ``batch_size``: Batch size for feature extraction.
            A larger batch size speeds up computation.
            You might have to experiment to find the largest batch size that fits in your GPU memory.
            See `here <https://github.com/krishnap25/mauve/issues/8#issuecomment-1082075240>`_ for details.

        :return: an object with fields p_hist, q_hist, divergence_curve and mauve.

        * ``out.mauve`` is a number between 0 and 1, the MAUVE score. Higher values means P is closer to Q.
        * ``out.frontier_integral``, a number between 0 and 1. Lower values mean that P is closer to Q.
        * ``out.p_hist`` is the obtained histogram for P. Same for ``out.q_hist``.
        * ``out.divergence_curve`` contains the points in the divergence curve. It is of shape (m, 2), where m is ``divergence_curve_discretization_size``

        """

        if p_features is None and p_tokens is None and p_text is None:
            raise ValueError('Supply at least one of p_features, p_tokens, p_text')
        if q_features is None and q_tokens is None and q_text is None:
            raise ValueError('Supply at least one of q_features, q_tokens, q_text')
        p_features = self.get_features_from_input(
            p_features, p_tokens, p_text, self.featurize_model_name, max_text_length,
            device_id, name="p", verbose=verbose, batch_size=batch_size, use_float64=use_float64,
        )
        q_features = self.get_features_from_input(
            q_features, q_tokens, q_text, self.featurize_model_name, max_text_length,
            device_id, name="q", verbose=verbose, batch_size=batch_size, use_float64=use_float64,
        )
        if num_buckets == 'auto':
            # heuristic: use num_clusters = num_generations / 10
            num_buckets = max(2, int(round(min(p_features.shape[0], q_features.shape[0]) / 10)))
        elif not isinstance(num_buckets, int):
            raise ValueError('num_buckets is expected to be an integer or "auto"')

        # Acutal binning
        t1 = time.time()
        p, q = self.cluster_feats(p_features, q_features,
                            num_clusters=num_buckets,
                            norm='l2', whiten=False,
                            pca_max_data=pca_max_data,
                            explained_variance=kmeans_explained_var,
                            num_redo=kmeans_num_redo,
                            max_iter=kmeans_max_iter,
                            seed=seed, verbose=verbose)
        t2 = time.time()
        if verbose:
            print('total discretization time:', round(t2-t1, 2), 'seconds')

        # Divergence curve and mauve
        mixture_weights = np.linspace(1e-6, 1-1e-6, divergence_curve_discretization_size)
        divergence_curve = self.get_divergence_curve_for_multinomials(p, q, mixture_weights, mauve_scaling_factor)
        x, y = divergence_curve.T
        idxs1 = np.argsort(x)
        idxs2 = np.argsort(y)
        mauve_score = 0.5 * (
            compute_area_under_curve(x[idxs1], y[idxs1]) +
            compute_area_under_curve(y[idxs2], x[idxs2])
        )
        fi_score = self.get_fronter_integral(p, q)
        to_return = SimpleNamespace(
            p_hist=p, q_hist=q, divergence_curve=divergence_curve,
            mauve=mauve_score,
            frontier_integral=fi_score,
            num_buckets=num_buckets,
        )
        return to_return

    def get_features_from_input(self, features, tokenized_texts, texts,
                                featurize_model_name, max_len, device_id, name, batch_size,
                                verbose=False, use_float64=False):
        global MODEL, TOKENIZER, MODEL_NAME
        if features is None:
            # Featurizing is necessary. Make sure the required packages are available
            if not FOUND_TORCH:
                raise ModuleNotFoundError(
                    """PyTorch not found. Please install PyTorch if you would like to use the featurization.
                        For details, see `https://github.com/krishnap25/mauve`
                        and `https://pytorch.org/get-started/locally/`.
                    """)
            if not FOUND_TRANSFORMERS:
                raise ModuleNotFoundError(
                    """Transformers not found. Please install Transformers if you would like to use the featurization.
                        For details, see `https://github.com/krishnap25/mauve`
                        and `https://huggingface.co/transformers/installation.html`.
                    """)

            if tokenized_texts is None:
                # tokenize texts
                if TOKENIZER is None or MODEL_NAME != featurize_model_name:
                    if verbose: print('Loading tokenizer')
                    TOKENIZER = get_tokenizer(featurize_model_name)
                if verbose: print('Tokenizing text...')
                tokenized_texts = [
                    TOKENIZER.encode(sen, return_tensors='pt', truncation=True, max_length=max_len)
                    for sen in texts
                ]
            # use tokenized_texts to featurize
            if TOKENIZER is None or MODEL_NAME != featurize_model_name:
                if verbose: print('Loading tokenizer')
                TOKENIZER = get_tokenizer(featurize_model_name)
            if MODEL is None or MODEL_NAME != featurize_model_name:
                if verbose: print('Loading model')
                MODEL = get_model(featurize_model_name, TOKENIZER, device_id)
                MODEL_NAME = featurize_model_name
            else:
                MODEL = MODEL.to(get_device_from_arg(device_id))
            if use_float64:
                MODEL = MODEL.double()
            if verbose: print('Featurizing tokens')
            features = featurize_tokens_from_model(MODEL, tokenized_texts, batch_size, name).detach().cpu().numpy()
        else:
            features = np.asarray(features)
        return features

    def cluster_feats(self, p, q, num_clusters,
                    norm='none', whiten=True,
                    pca_max_data=-1,
                    explained_variance=0.9,
                    num_redo=5, max_iter=500,
                    seed=0, verbose=False):
        assert 0 < explained_variance < 1
        if verbose:
            print(f'seed = {seed}')
        assert norm in ['none', 'l2', 'l1', None]
        data1 = np.vstack([q, p])
        if norm in ['l2', 'l1']:
            data1 = normalize(data1, norm=norm, axis=1)
        pca = PCA(n_components=None, whiten=whiten, random_state=seed+1)
        if pca_max_data < 0 or pca_max_data >= data1.shape[0]:
            pca.fit(data1)
        elif 0 < pca_max_data < data1.shape[0]:
            rng = np.random.RandomState(seed+5)
            idxs = rng.choice(data1.shape[0], size=pca_max_data, replace=False)
            pca.fit(data1[idxs])
        else:
            raise ValueError(f'Invalid argument pca_max_data={pca_max_data} with {data1.shape[0]} datapoints')
        s = np.cumsum(pca.explained_variance_ratio_)
        idx = np.argmax(s >= explained_variance)  # last index to consider
        if verbose:
            print(f'performing clustering in lower dimension = {idx}')
        data1 = pca.transform(data1)[:, :idx+1]
        # Cluster
        data1 = data1.astype(np.float32)
        t1 = time.time()
        kmeans = faiss.Kmeans(data1.shape[1], num_clusters, niter=max_iter,
                            verbose=verbose, nredo=num_redo, update_index=True,
                            seed=seed+2)
        kmeans.train(data1)
        _, labels = kmeans.index.search(data1, 1)
        labels = labels.reshape(-1)
        t2 = time.time()
        if verbose:
            print('kmeans time:', round(t2-t1, 2), 's')

        q_labels = labels[:len(q)]
        p_labels = labels[len(q):]

        q_bins = np.histogram(q_labels, bins=num_clusters,
                            range=[0, num_clusters], density=True)[0]
        p_bins = np.histogram(p_labels, bins=num_clusters,
                            range=[0, num_clusters], density=True)[0]
        return p_bins / p_bins.sum(), q_bins / q_bins.sum()


    def kl_multinomial(self, p, q):
        assert p.shape == q.shape
        if np.logical_and(p != 0, q == 0).any():
            return np.inf
        else:
            idxs = np.logical_and(p != 0, q != 0)
            return np.sum(p[idxs] * np.log(p[idxs] / q[idxs]))


    def get_divergence_curve_for_multinomials(self, p, q, mixture_weights, scaling_factor):
        # TODO: check if extreme points are needed
        divergence_curve = [[0, np.inf]] # extreme point
        for w in np.sort(mixture_weights):
            r = w * p + (1 - w) * q
            divergence_curve.append([self.kl_multinomial(q, r), self.kl_multinomial(p, r)])
        divergence_curve.append([np.inf, 0]) # other extreme point
        return np.exp(-scaling_factor * np.asarray(divergence_curve))

    def get_fronter_integral(self, p, q, scaling_factor=2):
        total = 0.0
        for p1, q1 in zip(p, q):
            if p1 == 0 and q1 == 0:
                pass
            elif p1 == 0:
                total += q1 / 4
            elif q1 == 0:
                total += p1 / 4
            elif abs(p1 - q1) > 1e-8:
                t1 = p1 + q1
                t2 = p1 * q1 * (math.log(p1) - math.log(q1)) / (p1 - q1)
                total += 0.25 * t1 - 0.5 * t2
            # else: contribution is 0
        return total * scaling_factor



"""
string2string code
src = https://github.com/stanfordnlp/string2string


MIT License

Copyright (c) 2023 Mirac Suzgun

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


"""

from typing import List, Union

# Take the Cartesian product of two lists of strings (or lists of lists of strings)
def cartesian_product(
    lst1: Union[List[str], List[List[str]]],
    lst2: Union[List[str], List[List[str]]],
    boolList: bool = False,
    list_of_list_separator: str = " ## ",
) -> Union[List[str], List[List[str]]]:
    """
    This function returns the Cartesian product of two lists of strings (or lists of lists of strings).

    Arguments:
        lst1: The first list of strings (or lists of lists of strings).
        lst2: The second list of strings (or lists of lists of strings).
        boolList: A boolean flag indicating whether the output should be a list of strings (or lists of lists of strings) (default: False).

    Returns:
        The Cartesian product of the two lists of strings (or lists of lists of strings).
    """
    if lst1 == []:
        return lst2
    elif lst2 == []:
        return lst1
    return [
        s1 + ("" if not (boolList) else list_of_list_separator) + s2
        for s1 in lst1
        for s2 in lst2
    ]